"""API for interacting with a Media DCR."""

from __future__ import annotations
from typing import Dict, List, Optional, TYPE_CHECKING, Union, Tuple
from typing_extensions import Self
from enum import Enum
import json
from ..archv2.release_policy_builder import ReleasePolicyBuilder
from ..archv2.release_policy_builder import ReleasePolicyBuilder
from .compute_job import (
    ComputeJob,
    MediaInsightsJob,
    MediaAudiencesValidationReportJob,
    MediaOverlapStatisticsJob,
    MediaDataAttributesJob,
    MediaLookalikeAudienceStatisticsJob,
    MediaGetAudiencesJob,
    MediaEstimateAudienceSizeJob,
    MediaAudienceUserListJob,
    MediaBaseAudienceValidationReportsJob,
    MediaModelQualityReportJob,
)
from .action import (
    CreateLookalikeAudienceAction,
    CreateRuleBasedAudienceAction,
    CreateRemarketingAudienceAction,
    UpdateCustomAudienceAction,
    DeleteCustomAudienceAction,
)
from .lookalike_audience_builder import LookalikeAudienceDefinition
from .rule_based_builder import RuleBasedAudienceDefinition
from .remarketing_audience_builder import RemarketingAudienceDefinition
from .participant import Participant
from .api import CustomAudience, SeedAudience
from ..types import MatchingId, MATCHING_ID_INTERNAL_LOOKUP
from ..archv2.client import DataRoomAuditLog
from .audiences import AudienceType

if TYPE_CHECKING:
    from ..client import Client


class DatasetType(Enum):
    """
    Enum for the type of dataset.

    Members:
        - AUDIENCES
        - MATCHING_IDS
        - DEMOGRAPHICS
        - SEGMENTS
        - EMBEDDINGS
    """

    AUDIENCES = "audiences"
    MATCHING_IDS = "matching"
    DEMOGRAPHICS = "demographics"
    SEGMENTS = "segments"
    EMBEDDINGS = "embeddings"


class CollaborationType(str, Enum):
    """
    Represents the features supported by a DCR.

    Members:
        - INSIGHTS
        - LOOKALIKE
        - REMARKETING
        - RULE_BASED
    """

    INSIGHTS = "ENABLE_INSIGHTS"
    LOOKALIKE = "ENABLE_LOOKALIKE_AUDIENCES"
    REMARKETING = "ENABLE_REMARKETING"
    RULE_BASED = "ENABLE_RULED_BASED_AUDIENCES"


class DcrConfiguration:
    def __init__(self, dcr_id: str, data_room: DataRoom):
        content = json.loads(data_room["signedContent"]["content"])
        self.config = {
            "id": dcr_id,
            "name": content["name"],
            "participants": content["participantGroups"],
            "matching_id_format": content["matchingIdFormat"],
            "matching_id_hashing_algorithm": content["hashMatchingIdWith"],
            "collaboration_types": [],
        }

        features = content["features"]
        if "ENABLE_INSIGHTS" in features:
            self.config["collaboration_types"].append(CollaborationType.INSIGHTS.value)
        if "ENABLE_LOOKALIKE_AUDIENCES" in features:
            self.config["collaboration_types"].append(CollaborationType.LOOKALIKE.value)
        if "ENABLE_REMARKETING" in features:
            self.config["collaboration_types"].append(
                CollaborationType.REMARKETING.value
            )
        if "ENABLE_RULED_BASED_AUDIENCES" in features:
            self.config["collaboration_types"].append(
                CollaborationType.RULE_BASED.value
            )

    def __str__(self):
        """
        Return a JSON string representation of the configuration.
        """
        return json.dumps(self.config, indent=2)

    # Added to mimic the behaviour of Pydantic types to provide a
    # consistent interface for the user.
    def model_dump_json(self):
        """
        Return a JSON string representation of the configuration.
        """
        return json.dumps(self.config, indent=2)


class MediaDcr:
    """
    Class for managing Media Data Clean Room (DCR) operations.
    """

    def __init__(
        self,
        client: Client,
        *,
        name: Optional[str] = None,
        participants: Optional[List[Participant]] = None,
        collaboration_types: Optional[List[CollaborationType]] = None,
        matching_ids: Optional[List[MatchingId]] = None,
        hide_absolute_values: bool = False,
        show_match_rate: bool = False,
        _existing_dcr_id: Optional[str] = None,
    ):
        """
        Initialize a MediaDCR instance.

        Can be used either to create a new DCR (when dcr_id is None) or load an existing one.

        **Parameters**:
        - `client`: The client instance for API communication
        - `name`: Name for new DCR
        - `participants`: Participants for new DCR
        - `collaboration_types`: Collaboration types for new DCR
        - `matching_ids`: Matching IDs for new DCR
        - `hide_absolute_values`: Whether to hide absolute values from outputs
        - `show_match_rate`: Whether to show match rate in outputs
        - `_existing_dcr_id`: Optional ID of existing DCR to load (for internal use only)
        """
        self.client = client
        self.client_v2 = client.client_v2
        self.session_v2 = self.client.create_session_v2()

        if _existing_dcr_id is not None:
            self.id = _existing_dcr_id
        else:
            if not all([name, participants, matching_ids]):
                raise Exception(
                    "name, participants, and matching_ids are required when creating a new DCR"
                )

            # Check that there is only one matching id for now.
            if len(matching_ids) == 1:
                self.matching_id = matching_ids[0]
            elif len(matching_ids) == 0:
                raise Exception("No matching id provided")
            else:
                raise Exception("Only one matching id is currently supported")

            self.id = self._build(
                name,
                participants,
                collaboration_types or [],
                matching_ids,
                hide_absolute_values,
                show_match_rate,
            )

    def _build(
        self,
        name: str,
        participants: List[Participant],
        collaboration_types: List[CollaborationType],
        matching_ids: List[MatchingId],
        hide_absolute_values: bool,
        show_match_rate: bool,
    ) -> str:
        """
        Build the Data Clean Room.
        """
        (
            matching_id_format,
            matching_id_hashing_algorithm,
        ) = MATCHING_ID_INTERNAL_LOOKUP[matching_ids[0]]
        create_data_room_definition = {
            "kind": "media",
            "audiencesPolicyId": None,
            "demographicsPolicyId": None,
            "embeddingsPolicyId": None,
            "enableDebugMode": False,
            "enableInsights": CollaborationType.INSIGHTS in collaboration_types,
            "enableLookalikeAudiences": CollaborationType.LOOKALIKE
            in collaboration_types,
            "enableRemarketing": CollaborationType.REMARKETING in collaboration_types,
            "enableDebugMode": False,
            "enableRuleBasedAudiences": CollaborationType.RULE_BASED
            in collaboration_types,
            "hashMatchingIdWith": matching_id_hashing_algorithm,
            "hideAbsoluteValuesFromInsights": hide_absolute_values,
            "showMatchRate": show_match_rate,
            "matchingIdFormat": matching_id_format,
            "matchingIdsPolicyId": None,
            "name": name,
            "participantGroups": [
                participant.as_dict() for participant in participants
            ],
            "policies": [],  # Create the DCR without policies
            "version": "v0",
            "fixOneHotEncodingColumnNameBug": True,
            "rawErrorsOutput": True,
            "useMatchingIdBasedOverlap": True,
        }
        verification_key = self.session_v2.send_get_verification_key_request()
        data_room = self.client_v2.create_data_room(
            create_data_room_definition, verification_key
        )
        return data_room["id"]

    def get_owner_email(self) -> str:
        """
        Get the email of the owner of the DCR.
        """
        # Get the creator of the DCR.
        data_room = self._get_data_room()
        owner_id = data_room["ownerId"]
        owner = self.client._graphql.post(
            """
            query GetUserEmail($id: ID!) {
                user(id: $id) {
                    email
                }
            }""",
            {"id": owner_id},
        )
        return owner["user"]["email"]

    def _get_data_room(self) -> DataRoom:
        """
        Get the data room.
        """
        verification_key = self.session_v2.send_get_verification_key_request()
        return self.client_v2.get_data_room(self.id, verification_key)

    def _get_policies(self) -> Dict[str, str]:
        """
        Get a map of leaf node names to policy ids.
        """
        data_room = self._get_data_room()
        dcr_state = json.loads(data_room["signedContent"]["content"])
        policies = {p["leafId"]: p["policyId"] for p in dcr_state["policies"]}
        return policies

    def provision_base_audience(self, data_lab_id: str) -> None:
        """
        Provision a base audience to the DCR.

        **Parameters**:
        - `data_lab_id`: The identifier of the DataLab from which to provision the base audience

        **Raises**:
        - `Exception`: If the DataLab is not validated or if provisioning fails
        """
        # Deprovision all existing datasets before provisioning new ones.
        # This ensures we don't end up with a mix of old and new datasets.
        self.deprovision_base_audience()

        # Check DataLab is validated
        data_lab = self.client.get_data_lab(data_lab_id)
        if not data_lab["isValidated"]:
            raise Exception("Cannot provision DataLab, not validated.")

        # Check compatibility
        data_lab_compatibility = self.client_v2.check_data_labs_compatibility(
            self.id, [data_lab_id]
        )
        error_messages = data_lab_compatibility["errorsByDataLab"][data_lab_id]
        if error_messages:
            messages = [e["message"] for e in error_messages]
            formatted_error_messages = ", ".join(messages)
            raise Exception(
                f"Cannot provision DataLab, incompatible. Error messages: {formatted_error_messages}"
            )

        dataset_manifest_hashes = [
            (d["name"], d["dataset"]["manifestHash"])
            for d in data_lab["datasets"]
            if d.get("dataset") is not None
        ]

        self._provision_base_audience(dataset_manifest_hashes)

    def _provision_base_audience(self, dataset_manifest_hashes: list[Tuple[str, str]]):
        dataset_policy_ids = {
            "MATCHING_DATA": None,
            "SEGMENTS_DATA": None,
            "DEMOGRAPHICS_DATA": None,
            "EMBEDDINGS_DATA": None,
        }
        for dataset_name, manifest_hash in dataset_manifest_hashes:
            dataset_type = None
            if dataset_name == "MATCHING_DATA":
                dataset_type = DatasetType.MATCHING_IDS
            elif dataset_name == "SEGMENTS_DATA":
                dataset_type = DatasetType.SEGMENTS
            elif dataset_name == "DEMOGRAPHICS_DATA":
                dataset_type = DatasetType.DEMOGRAPHICS
            elif dataset_name == "EMBEDDINGS_DATA":
                dataset_type = DatasetType.EMBEDDINGS
            else:
                raise ValueError(f"Unknown datalab dataset name: {dataset_name}")

            policy_id = self._create_release_policy(
                manifest_hash,
                dataset_type,
            )
            dataset_policy_ids[dataset_name] = policy_id

        # Provision datasets from DataLab to DCR
        provision_action = {
            "kind": "media",
            "provisionBaseAudiencePolicyIds": {
                "demographicsPolicyId": dataset_policy_ids.get("DEMOGRAPHICS_DATA"),
                "embeddingsPolicyId": dataset_policy_ids.get("EMBEDDINGS_DATA"),
                "matchingPolicyId": dataset_policy_ids.get("MATCHING_DATA"),
                "segmentsPolicyId": dataset_policy_ids.get("SEGMENTS_DATA"),
            },
        }
        self.session_v2.send_data_room_state_action_request(self.id, provision_action)

    def is_base_audience_provisioned(self) -> bool:
        """
        Check if the base audience is provisioned.
        """
        policies = self._get_policies()
        # Check that the matching policy is provisioned.
        # This will indicate that the base audience is provisioned.
        return "matching" in policies and policies["matching"] is not None

    def deprovision_base_audience(self) -> None:
        """
        Deprovision a base audience from the DCR.

        Removes all datasets associated with the base audience from the DCR by deprovisioning
        each dataset type.
        """
        deprovision_action = {"kind": "media", "deprovisionBaseAudiencePolicyIds": {}}
        self.session_v2.send_data_room_state_action_request(self.id, deprovision_action)

    def provision_seed_audiences(self, manifest_hash: str) -> None:
        """
        Provision seed audiences dataset to the DCR.

        **Parameters**:
        - `manifest_hash`: Hash of the seed audience dataset to provision
        """
        policy_id = self._create_release_policy(manifest_hash, DatasetType.AUDIENCES)
        dataset_action = {
            "kind": "media",
            "provisionSeedAudiencePolicyIds": {"audiencesPolicyId": policy_id},
        }
        self.session_v2.send_data_room_state_action_request(self.id, dataset_action)

    def are_seed_audiences_provisioned(self) -> bool:
        """
        Check if the seed audiences are provisioned.
        """
        policies = self._get_policies()
        return "audiences" in policies and policies["audiences"] is not None

    def deprovision_seed_audiences(self) -> None:
        """Deprovision seed audiences dataset from the DCR."""
        # Set policy_id to None to indicate that the dataset is no longer used.
        dataset_action = {"kind": "media", "deprovisionSeedAudiencePolicyIds": {}}
        self.session_v2.send_data_room_state_action_request(self.id, dataset_action)

    def create_lookalike_audience(
        self,
        definition: LookalikeAudienceDefinition,
    ) -> CustomAudience:
        """
        Create a lookalike audience based on the provided definition.

        **Parameters**:
        - `definition`: Definition of the lookalike audience to create

        **Returns**:
        - The created lookalike audience
        """
        action = CreateLookalikeAudienceAction(
            dcr_id=self.id,
            client=self.client,
            definition=definition,
        )
        response = action.create()
        lal_audience = CustomAudience.model_validate(
            response["createCustomAudience"]["audience"]
        )
        return lal_audience

    def create_rule_based_audience(
        self,
        definition: RuleBasedAudienceDefinition,
    ) -> CustomAudience:
        """
        Create a rule-based audience based on the provided definition.

        **Parameters**:
        - `definition`: Definition of the rule-based audience to create

        **Returns**:
        - The created rule-based audience
        """
        action = CreateRuleBasedAudienceAction(
            dcr_id=self.id,
            client=self.client,
            definition=definition,
        )
        response = action.create()
        rb_audience = CustomAudience.model_validate(
            response["createCustomAudience"]["audience"]
        )
        return rb_audience

    def create_remarketing_audience(
        self,
        definition: RemarketingAudienceDefinition,
    ) -> CustomAudience:
        """
        Create a remarketing audience based on the provided definition.

        **Parameters**:
        - `definition`: Definition of the remarketing audience to create

        **Returns**:
        - The created remarketing audience
        """
        action = CreateRemarketingAudienceAction(
            dcr_id=self.id,
            client=self.client,
            definition=definition,
        )
        response = action.create()
        rm_audience = CustomAudience.model_validate(
            response["createCustomAudience"]["audience"]
        )
        return rm_audience

    def get_insights(self) -> MediaInsightsJob:
        """
        Compute insights from the data room.

        **Returns**:
        - The job to compute insights
        """
        job = MediaInsightsJob(self.id, self.client)
        return self._run_compute_job(job)

    def get_base_audience_validation_reports(
        self,
    ) -> MediaBaseAudienceValidationReportsJob:
        """
        Get the validation reports for the DCR.

        **Returns**:
        - The job to compute the validation reports
        """
        job = MediaBaseAudienceValidationReportsJob(self.id, self.client)
        return self._run_compute_job(job)

    def get_seed_audiences_validation_report(
        self,
    ) -> MediaAudiencesValidationReportJob:
        """
        Get the validation report for seed audiences data.

        **Returns**:
        - The job to compute the seed audiences validation report
        """
        job = MediaAudiencesValidationReportJob(self.id, self.client)
        return self._run_compute_job(job)

    def get_overlap_statistics(self) -> MediaOverlapStatisticsJob:
        """
        Compute overlap statistics.

        **Returns**:
        - The job to compute the overlap statistics
        """
        job = MediaOverlapStatisticsJob(self.id, self.client)
        return self._run_compute_job(job)

    def get_data_attributes(self) -> MediaDataAttributesJob:
        """
        Get data attributes.

        **Returns**:
        - The job to get the data attributes
        """
        job = MediaDataAttributesJob(self.id, self.client)
        return self._run_compute_job(job)

    def get_lookalike_audience_statistics(
        self,
        audience: AudienceType,
    ) -> MediaLookalikeAudienceStatisticsJob:
        """
        Get statistics for lookalike audiences.

        **Returns**:
        - The job to get the lookalike audience statistics
        """
        job = MediaLookalikeAudienceStatisticsJob(self.id, audience, self.client)
        return self._run_compute_job(job)

    def get_audiences(
        self,
    ) -> MediaGetAudiencesJob:
        """
        Get audiences for a specific user.

        **Returns**:
        - The job to get the audiences for the specified user
        """
        participants = self.get_participants()
        user_participant_id = None
        for participant in participants:
            if self.client.user_email in participant.emails:
                user_participant_id = participant.id
                break
        if user_participant_id is None:
            raise Exception(
                f"User {self.client.user_email} does not appear in any of the participants"
            )
        else:
            job = MediaGetAudiencesJob(
                self.id,
                self.client.user_email,
                self.client,
                participant_id=user_participant_id,
            )
            return self._run_compute_job(job)

    def share_audiences(
        self, audiences: List[CustomAudience], participants: List[Participant]
    ):
        """
        Share the audiences with the participant group.

        **Parameters**:
        - `audiences`: The audiences to share
        - `participants`: The participants to share the audiences with
        """
        for audience in audiences:
            audience.sharedWith = [participant.id for participant in participants]
            action = UpdateCustomAudienceAction(self.id, self.client, audience)
            action.update()

    def delete_audiences(
        self,
        audiences: List[CustomAudience],
        *,
        delete_dependent_audiences: bool = False,
    ) -> None:
        """
        Delete audiences.

        **Parameters**:
        - `audiences`: The audiences to delete
        - `delete_dependent_audiences`: If true, the supplied audiences and all dependent audiences will be deleted. If false, the supplied audiences will only be deleted if they have no dependents.
        """
        for audience in audiences:
            # Delete the user requested audiences
            action = DeleteCustomAudienceAction(
                self.id, self.client, audience.id, delete_dependent_audiences
            )
            action.delete()

    def get_audience_size(
        self, audience: Union[SeedAudience, CustomAudience]
    ) -> MediaEstimateAudienceSizeJob:
        """
        Get the size of an audience.

        **Parameters**:
        - `audience`: The audience to estimate the size of

        **Returns**:
        - The job to get the size of the audience
        """
        job = MediaEstimateAudienceSizeJob(self.id, audience, self.client)
        return self._run_compute_job(job)

    def get_audience_user_list(
        self, audience: Union[SeedAudience, CustomAudience]
    ) -> MediaAudienceUserListJob:
        """
        Get the user list for an audience.

        **Parameters**:
        - `audience`: The audience to get the user list for

        **Returns**:
        - The job to get the user list for the audience
        """
        job = MediaAudienceUserListJob(self.id, audience, self.client)
        return self._run_compute_job(job)

    def get_audit_log(self) -> List[DataRoomAuditLog]:
        """
        Get the audit log for the data room.

        **Returns**:
        - Audit log
        """
        verification_key = self.session_v2.send_get_verification_key_request()
        verified_audit_log = self.client_v2.get_data_room_audit_log(
            self.id, verification_key
        )
        return verified_audit_log

    def _create_release_policy(
        self, manifest_hash: str, dataset_type: DatasetType
    ) -> str:
        """
        Create a release policy for a dataset.

        **Parameters**:
        - `manifest_hash`: The hash of the dataset
        - `dataset_type`: The type of the dataset

        **Returns**:
        - The id of the created release policy
        """
        release_policy = (
            ReleasePolicyBuilder(self.client, self.id, manifest_hash)
            .with_target_leaf_node_name(dataset_type.value)
            .build()
        )
        policy_id = self.session_v2.create_policy(release_policy)
        return policy_id

    @classmethod
    def from_existing(cls, dcr_id: str, client: Client) -> Self:
        """
        Construct a MediaDCR from an existing DCR with the given ID.

        **Parameters**:
        - `dcr_id`: The id of the DCR
        - `client`: The client from which a session is created

        **Returns**:
        - The MediaDCR instance
        """
        return cls(client=client, _existing_dcr_id=dcr_id)

    def get_participants(self) -> List[Participant]:
        """
        Get the participants of the DCR.

        **Returns**:
        - The list of participants in the DCR
        """
        verification_key = self.session_v2.send_get_verification_key_request()
        dr = self.client_v2.get_data_room(self.id, verification_key)
        content = json.loads(dr["signedContent"]["content"])
        existing_participants = content["participantGroups"]
        return [
            Participant(
                emails=pg["emails"],
                permissions=pg["permissions"],
                role=pg["role"],
                id=pg["id"],
                organization_id=pg["organizationId"],
            )
            for pg in existing_participants
            if "model.quality@decentriq.com" not in pg["emails"]
        ]

    def can_other_participants_export_audience(self) -> bool:
        """
        Check if other participants can export audiences.
        """
        other_participants = [
            participant
            for participant in self.get_participants()
            if self.client.user_email not in participant.emails
        ]
        return any(
            participant.can_export_audience() for participant in other_participants
        )

    def get_configuration(self) -> DcrConfiguration:
        """
        Get the configuration of the DCR.

        **Returns**:
        - The configuration of the DCR
        """
        verification_key = self.session_v2.send_get_verification_key_request()
        dr = self.client_v2.get_data_room(self.id, verification_key)
        return DcrConfiguration(self.id, dr)

    def _run_compute_job(self, compute_job: ComputeJob) -> ComputeJob:
        """
        Run a compute job and return the result.

        **Parameters**:
        - `compute_job`: The compute job to run

        **Returns**:
        - The compute job
        """
        compute_job.run()
        return compute_job

    def _retrieve_model_quality_report(self) -> MediaModelQualityReportJob:
        job = MediaModelQualityReportJob(self.id, self.client)
        return self._run_compute_job(job)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: identity_endorsement.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'identity_endorsement.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1aidentity_endorsement.proto\x12\x14identity_endorsement\"\xd2\x01\n\x12\x45ndorsementRequest\x12L\n\x15pkiEndorsementRequest\x18\x02 \x01(\x0b\x32+.identity_endorsement.PkiEndorsementRequestH\x00\x12X\n\x1b\x64\x63rSecretEndorsementRequest\x18\x03 \x01(\x0b\x32\x31.identity_endorsement.DcrSecretEndorsementRequestH\x00\x42\x14\n\x12\x65ndorsementRequest\"\xd8\x01\n\x13\x45ndorsementResponse\x12N\n\x16pkiEndorsementResponse\x18\x02 \x01(\x0b\x32,.identity_endorsement.PkiEndorsementResponseH\x00\x12Z\n\x1c\x64\x63rSecretEndorsementResponse\x18\x03 \x01(\x0b\x32\x32.identity_endorsement.DcrSecretEndorsementResponseH\x00\x42\x15\n\x13\x65ndorsementResponse\"4\n\x15PkiEndorsementRequest\x12\x1b\n\x13\x63\x65rtificateChainPem\x18\x01 \x01(\x0c\"Z\n\x16PkiEndorsementResponse\x12@\n\x0epkiEndorsement\x18\x01 \x01(\x0b\x32(.identity_endorsement.EnclaveEndorsement\"0\n\x1b\x44\x63rSecretEndorsementRequest\x12\x11\n\tdcrSecret\x18\x01 \x01(\t\"{\n\x1c\x44\x63rSecretEndorsementResponse\x12\x46\n\x14\x64\x63rSecretEndorsement\x18\x01 \x01(\x0b\x32(.identity_endorsement.EnclaveEndorsement\x12\x13\n\x0b\x64\x63rSecretId\x18\x02 \x01(\x0c\"\xca\x01\n\x13\x45nclaveEndorsements\x12=\n\x0bpersonalPki\x18\x01 \x01(\x0b\x32(.identity_endorsement.EnclaveEndorsement\x12\x37\n\x05\x64qPki\x18\x02 \x01(\x0b\x32(.identity_endorsement.EnclaveEndorsement\x12;\n\tdcrSecret\x18\x03 \x01(\x0b\x32(.identity_endorsement.EnclaveEndorsement\"7\n\x12\x45nclaveEndorsement\x12!\n\x19\x65ndorsementCertificateDer\x18\x01 \x01(\x0c\"&\n\x08PkiClaim\x12\x1a\n\x12rootCertificateDer\x18\x01 \x01(\x0c\"%\n\x0e\x44\x63rSecretClaim\x12\x13\n\x0b\x64\x63rSecretId\x18\x01 \x01(\x0c\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'identity_endorsement_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ENDORSEMENTREQUEST']._serialized_start=53
  _globals['_ENDORSEMENTREQUEST']._serialized_end=263
  _globals['_ENDORSEMENTRESPONSE']._serialized_start=266
  _globals['_ENDORSEMENTRESPONSE']._serialized_end=482
  _globals['_PKIENDORSEMENTREQUEST']._serialized_start=484
  _globals['_PKIENDORSEMENTREQUEST']._serialized_end=536
  _globals['_PKIENDORSEMENTRESPONSE']._serialized_start=538
  _globals['_PKIENDORSEMENTRESPONSE']._serialized_end=628
  _globals['_DCRSECRETENDORSEMENTREQUEST']._serialized_start=630
  _globals['_DCRSECRETENDORSEMENTREQUEST']._serialized_end=678
  _globals['_DCRSECRETENDORSEMENTRESPONSE']._serialized_start=680
  _globals['_DCRSECRETENDORSEMENTRESPONSE']._serialized_end=803
  _globals['_ENCLAVEENDORSEMENTS']._serialized_start=806
  _globals['_ENCLAVEENDORSEMENTS']._serialized_end=1008
  _globals['_ENCLAVEENDORSEMENT']._serialized_start=1010
  _globals['_ENCLAVEENDORSEMENT']._serialized_end=1065
  _globals['_PKICLAIM']._serialized_start=1067
  _globals['_PKICLAIM']._serialized_end=1105
  _globals['_DCRSECRETCLAIM']._serialized_start=1107
  _globals['_DCRSECRETCLAIM']._serialized_end=1144
# @@protoc_insertion_point(module_scope)



from enum import Enum
from typing import List, Literal, Optional, Union

from ...proto import ComputeNodeFormat
from ...proto.length_delimited import serialize_length_delimited
from ..node import Node
from .proto.compute_s3_sink_pb2 import (
    S3Object,
    S3Provider,
    S3SinkWorkerConfiguration,
    DqDspCredentials,
    UserDefinedCredentials,
    DspCredentialsType,
)


class DspType(Enum):
    SPLICKY = 0


class S3SinkCompute(Node):
    """
    Compute node which allows the upload of files to an S3 api compatible endpoint
    """

    def __init__(
        self,
        name: str,
        endpoint: str,
        credentials_dependency: Union[str, DspType],
        s3_provider: Literal["AWS", "GCS"],
        objects: List[S3Object],
        region: str = "",
        content_type: Optional[str] = None,
    ) -> None:
        """
        Create a container compute node.

        **Parameters**:
        - `name`: The name of the node. This serves as an identifier for the
            node and needs to be specified when you interact with the node
            (e.g. run its computation or retrieve its results).
        - `endpoint`: The S3 endpoint including the bucket name.
        - `region`: The S3 region used by the endpoint.
        - `credentials_dependency`: the id of the dependency that holds the credentials
            to access the S3 bucket
            This node should provide a single JSON with format:
            - For AWS:
            ```
            {
                "accessKey": "xxxx",
                "secretKey": "yyyy"
            }
            - For GCS, see [service account credentials](https://cloud.google.com/iam/docs/service-account-creds)
        - `objects`: The list of objects to upload to the S3 bucket
        """
        user_defined_credentials = None
        dq_dsp_credentials = None
        if isinstance(credentials_dependency, str):
            user_defined_credentials = UserDefinedCredentials(
                userDefinedCredentials=credentials_dependency
            )
            dependencies = [credentials_dependency] + list(
                map(lambda a: a.dependency, objects)
            )
        else:
            if credentials_dependency == DspType.SPLICKY:
                dq_dsp_credentials = DqDspCredentials(dspType=DspCredentialsType.SPLICKY)
            else:
                raise Exception("Unknown DSP credentials")
            dependencies = list(map(lambda a: a.dependency, objects))

        configuration = S3SinkWorkerConfiguration(
            endpoint=endpoint,
            region=region,
            userDefinedCredentials=user_defined_credentials,
            dqDspCredentials=dq_dsp_credentials,
            contentType=content_type,
        )

        configuration.objects.extend(objects)
        if s3_provider == "AWS":
            configuration.s3Provider = S3Provider.AWS
        elif s3_provider == "GCS":
            configuration.s3Provider = S3Provider.GCS
        else:
            raise Exception(f"Invalid s3_provider: got {s3_provider}")

        config = serialize_length_delimited(configuration)

        super().__init__(
            name,
            config=config,
            enclave_type="decentriq.s3-sink-worker",
            dependencies=dependencies,
            output_format=ComputeNodeFormat.RAW,
        )


from ...decoder import S3SinkWorkerDecoder



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: compute_sql.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'compute_sql.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11\x63ompute_sql.proto\x12\x0b\x63ompute_sql\"\xa3\x01\n\x16SqlWorkerConfiguration\x12:\n\nvalidation\x18\x01 \x01(\x0b\x32$.compute_sql.ValidationConfigurationH\x00\x12<\n\x0b\x63omputation\x18\x02 \x01(\x0b\x32%.compute_sql.ComputationConfigurationH\x00\x42\x0f\n\rconfiguration\"H\n\x17ValidationConfiguration\x12-\n\x0btableSchema\x18\x01 \x01(\x0b\x32\x18.compute_sql.TableSchema\"=\n\x0bTableSchema\x12.\n\x0cnamedColumns\x18\x01 \x03(\x0b\x32\x18.compute_sql.NamedColumn\"V\n\x0bNamedColumn\x12\x11\n\x04name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12+\n\ncolumnType\x18\x02 \x01(\x0b\x32\x17.compute_sql.ColumnTypeB\x07\n\x05_name\"Q\n\nColumnType\x12\x31\n\rprimitiveType\x18\x01 \x01(\x0e\x32\x1a.compute_sql.PrimitiveType\x12\x10\n\x08nullable\x18\x02 \x01(\x08\";\n\x16TableDependencyMapping\x12\r\n\x05table\x18\x01 \x01(\t\x12\x12\n\ndependency\x18\x02 \x01(\t\"\xdb\x01\n\x18\x43omputationConfiguration\x12\x14\n\x0csqlStatement\x18\x01 \x01(\t\x12\x35\n\x0fprivacySettings\x18\x02 \x01(\x0b\x32\x1c.compute_sql.PrivacySettings\x12,\n\x0b\x63onstraints\x18\x03 \x03(\x0b\x32\x17.compute_sql.Constraint\x12\x44\n\x17tableDependencyMappings\x18\x04 \x03(\x0b\x32#.compute_sql.TableDependencyMapping\"2\n\x0fPrivacySettings\x12\x1f\n\x17minAggregationGroupSize\x18\x01 \x01(\x03\"!\n\nConstraint\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t*3\n\rPrimitiveType\x12\t\n\x05INT64\x10\x00\x12\n\n\x06STRING\x10\x01\x12\x0b\n\x07\x46LOAT64\x10\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'compute_sql_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PRIMITIVETYPE']._serialized_start=878
  _globals['_PRIMITIVETYPE']._serialized_end=929
  _globals['_SQLWORKERCONFIGURATION']._serialized_start=35
  _globals['_SQLWORKERCONFIGURATION']._serialized_end=198
  _globals['_VALIDATIONCONFIGURATION']._serialized_start=200
  _globals['_VALIDATIONCONFIGURATION']._serialized_end=272
  _globals['_TABLESCHEMA']._serialized_start=274
  _globals['_TABLESCHEMA']._serialized_end=335
  _globals['_NAMEDCOLUMN']._serialized_start=337
  _globals['_NAMEDCOLUMN']._serialized_end=423
  _globals['_COLUMNTYPE']._serialized_start=425
  _globals['_COLUMNTYPE']._serialized_end=506
  _globals['_TABLEDEPENDENCYMAPPING']._serialized_start=508
  _globals['_TABLEDEPENDENCYMAPPING']._serialized_end=567
  _globals['_COMPUTATIONCONFIGURATION']._serialized_start=570
  _globals['_COMPUTATIONCONFIGURATION']._serialized_end=789
  _globals['_PRIVACYSETTINGS']._serialized_start=791
  _globals['_PRIVACYSETTINGS']._serialized_end=841
  _globals['_CONSTRAINT']._serialized_start=843
  _globals['_CONSTRAINT']._serialized_end=876
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import json

from typing import TYPE_CHECKING, Dict, Any, Optional, List
from typing_extensions import Self
from decentriq_dcr_compiler import ab_media as ab_media_compiler

from ..session import Session
from ..attestation import enclave_specifications
from ..types import EnclaveSpecification
from .advertiser_api import AdvertiserApi
from .publisher_api import PublisherApi
from .features import AbMediaDcrFeatures
from .version import AUDIENCE_BUILDER_WRAPPER_SUPPORTED_VERSION

if TYPE_CHECKING:
    from ..client import Client

__docformat__ = "restructuredtext"


class AbMediaDcrDefinition:
    """
    Class representing an Audience Builder DCR Definition.
    """

    def __init__(
        self,
        name: str,
        high_level: Dict[str, Any],
        enclave_specs: Optional[Dict[str, EnclaveSpecification]] = None,
    ) -> None:
        self.name = name
        self._high_level = high_level
        self._enclave_specs = enclave_specs


class AbMediaDcr:
    """
    Class representing an Audience Builder DCR.
    """

    def __init__(
        self,
        dcr_id: str,
        high_level: Dict[str, Any],
        session: Session,
        *,
        client: Client,
    ) -> None:
        """
        Initialise an Audience Builder DCR.

        **Parameters**:
        - `dcr_id`: ID of the Audience Builder DCR.
        - `high_level`: High level representation of the Audience Builder DCR.
        - `session`: A `Session` object which can be used for communication with the enclave.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        """
        self.client = client
        self.session = session
        self.high_level = high_level
        self.id = dcr_id
        self.features = _get_features(high_level)
        self.advertiser = AdvertiserApi(self)
        self.publisher = PublisherApi(self)

    def retrieve_audit_log(self) -> str:
        """
        Retrieve the audit log.
        """
        return self.session.retrieve_audit_log(self.id).log.decode("utf-8")

    def participants(self) -> Dict[str, Any]:
        """
        Retrieve the participants of the Audience Builder DCR.
        This returns a dictionary of roles (keys) mapped to participants (email addresses).
        """
        dcr = self.high_level[AUDIENCE_BUILDER_WRAPPER_SUPPORTED_VERSION]
        compute_keys_list = list(dcr["compute"].keys())
        if len(compute_keys_list) != 1:
            raise Exception(
                f"Unable to extract Media DCR version. Expected a single top-level property indicating the DCR version."
            )

        compute_version = compute_keys_list[0]
        compute = dcr["compute"][compute_version]
        return {
            "publisher": compute["publisherEmails"],
            "advertiser": compute["advertiserEmails"],
            "observer": compute["observerEmails"],
            "agency": compute["agencyEmails"],
            "data_partner": compute["dataPartnerEmails"],
        }

    def stop(self):
        """
        Stop the Audience Builder DCR.
        """
        self.session.stop_data_room(self.id)

    @classmethod
    def _from_existing(
        cls,
        dcr_id: str,
        *,
        client: Client,
        enclave_specs: Optional[List[EnclaveSpecification]] = None,
    ) -> Self:
        """
        Construct an Audience Builder DCR from an existing DCR with the given ID.

        **Parameters**:
        - `dcr_id`: ID of the Audience Builder DCR.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `enclave_specs`: Determines the types of enclaves that are supported by this Data Clean Room.
            If not specified, the latest enclave specifications will be used.
        """
        specs = enclave_specs if enclave_specs else enclave_specifications.all()
        if enclave_specs:
            specs_dict = {spec["name"]: spec for spec in specs}
        else:
            specs_dict = None
        existing_data_room_description = client.get_data_room_description(
            dcr_id, enclave_specs=specs_dict
        )
        if not existing_data_room_description:
            raise Exception(
                f"Unable to retrieve data room description for data room with ID {dcr_id}"
            )
        session = client.create_session_from_data_room_description(
            existing_data_room_description, specs
        )
        existing_dcr = session.retrieve_data_room(dcr_id)
        high_level = json.loads(existing_dcr.highLevelRepresentation.decode())

        dcr = cls(dcr_id=dcr_id, session=session, high_level=high_level, client=client)
        return dcr


def _get_features(high_level: Dict[str, Any]) -> AbMediaDcrFeatures:
    features = ab_media_compiler.get_ab_media_dcr_features_serialized(
        json.dumps(high_level)
    )
    return AbMediaDcrFeatures(features)



from __future__ import annotations
import json
from typing import (
    TYPE_CHECKING,
    Any,
    Tuple,
    cast
)

from ..types import JSONType
from ..connection import Connection
from ..channel import EnclaveError
from ..proto import (
    GcgRequestV2,
    AuthenticatedRequest,
    AuthenticatedResponse,
    DcrActionRequest,
    CreatePolicyRequest,
    CreatePolicyResponse,
    ExportResultAsDatasetRequest,
    RawExport,
    RetrieveResultEncryptionKeyRequest,
    RetrieveResultEncryptionKeyResponse,
    ReleasePolicy,
    GetVerificationKeyRequest,
)
from ..proto.secret_store_pb2 import (
    SecretStoreEntry,
    SecretStoreRequest,
    SecretStoreResponse,
    GetSecretRequest,
    RemoveSecretRequest,
    CreateSecretRequest,
    UpdateSecretAclRequest,
)
from .secret import Secret

from decentriq_dcr_compiler.schemas.secret_store_entry_state import SecretStoreEntryState, SECRET_STORE_ENTRY_STATE_VERSION, v0
from decentriq_dcr_compiler.schemas.data_room import DataRoomAction as DataRoomActionSchema

if TYPE_CHECKING:
    from ..client import Client

__all__ = [
    "SessionV2",
]

class SessionV2:
    """
    Class for managing the communication with an enclave.
    """

    client: Client
    connection: Connection
    keypair: Any

    def __init__(
        self,
        client: Client,
        connection: Connection,
    ):
        """
        `Session` instances should not be instantiated directly but rather
         be created using a `Client` object using  `decentriq_platform.Client.create_session_v2`.
        """
        self.client = client
        self.connection = connection

    def send_authenticated_request(
        self,
        authenticated_request: AuthenticatedRequest,
    ) -> AuthenticatedResponse:
        authenticated_request.apiToken = self.client.enclave_api_token
        request = GcgRequestV2(authenticated=authenticated_request)
        response = self.connection.send_request_v2(request)
        if response.HasField("failure"):
            raise EnclaveError(response.failure)
        successful_response = response.success
        if not successful_response.HasField("authenticated"):
            raise Exception(
                "Expected `authenticated` response, got "
                + str(successful_response.WhichOneof("response"))
            )
        authenticated_response = response.success.authenticated
        return authenticated_response

    def send_secret_store_request(
        self,
        request: SecretStoreRequest,
    ) -> SecretStoreResponse:
        authenticated_request = AuthenticatedRequest(secretStore=request)
        response = self.send_authenticated_request(authenticated_request)
        if not response.HasField("secretStore"):
            raise Exception(
                f"Expected `secretStore` response, got "
                + str(response.WhichOneof("response"))
            )
        secret_store_response = cast(SecretStoreResponse, response.secretStore)
        return secret_store_response

    def remove_secret(self, secret_id: str, expected_cas_index: int) -> bool:
        request = SecretStoreRequest(
            removeSecret=RemoveSecretRequest(
                id=secret_id,
                expectedCasIndex=expected_cas_index,
            )
        )
        secret_store_response = self.send_secret_store_request(request)
        if not secret_store_response.HasField("removeSecret"):
            raise Exception(
                f"Expected `removeSecret`, got "
                + str(secret_store_response.WhichOneof("response"))
            )
        return secret_store_response.removeSecret.removed

    def get_secret(self, secret_id: str) -> Tuple[Secret, int]:
        request = SecretStoreRequest(
            getSecret=GetSecretRequest(
                id=secret_id,
                version=SECRET_STORE_ENTRY_STATE_VERSION
            )
        )
        secret_store_response = self.send_secret_store_request(request)
        if not secret_store_response.HasField("getSecret"):
            raise Exception(
                f"Expected `getSecret`, got "
                + str(secret_store_response.WhichOneof("response"))
            )
        get_secret_response = secret_store_response.getSecret
        if not get_secret_response.HasField("secret"):
            raise Exception(
                f"Expected `secret` in `getSecret` response, got "
                + str(get_secret_response)
            )
        else:
            entry_state = SecretStoreEntryState.model_validate_json(get_secret_response.secret.state)
            return Secret(get_secret_response.secret.content, entry_state),  get_secret_response.casIndex

    def create_secret(self, secret: Secret) -> str:
        """Store a secret in the user's own enclave-protected secret store"""
        request = SecretStoreRequest(
            createSecret=CreateSecretRequest(
                secret=SecretStoreEntry(
                    content=secret.secret,
                    state=secret.state.model_dump_json().encode("utf-8")
                )
            )
        )
        secret_store_response = self.send_secret_store_request(request)
        if not secret_store_response.HasField("createSecret"):
            raise Exception(
                f"Expected `setSecret`, got "
                + str(secret_store_response.WhichOneof("response"))
            )
        return secret_store_response.createSecret.id

    def get_dataset_secret_id(self, manifest_hash: str) -> str:
        pass

    def update_secret_acl(self, secret_id: str, new_acl: v0.SecretStoreEntryAcl, expected_cas_index: int) -> bool:
        """Update a secret ACL"""
        request = SecretStoreRequest(
            updateSecretAcl=UpdateSecretAclRequest(
                id=secret_id,
                newAcl=new_acl.model_dump_json().encode("utf-8"),
                version=SECRET_STORE_ENTRY_STATE_VERSION,
                expectedCasIndex=expected_cas_index
            )
        )
        secret_store_response = self.send_secret_store_request(request)
        if not secret_store_response.HasField("updateSecretAcl"):
            raise Exception(
                f"Expected `updatedSecretAcl`, got "
                + str(secret_store_response.WhichOneof("response"))
            )
        return secret_store_response.updateSecretAcl.updated

    def create_policy(self, policy: ReleasePolicy) -> str:
        """Create a release policy and return the policy ID."""
        scope_id = self.client._ensure_dataset_scope()
        scope_id_bytes = bytes.fromhex(scope_id)

        request = CreatePolicyRequest(
            policy=policy,
            scope=scope_id_bytes
        )
        authenticated_request = AuthenticatedRequest(createPolicy=request)
        response = self.send_authenticated_request(authenticated_request)
        if not response.HasField("createPolicy"):
            raise Exception(
                f"Expected `createPolicy` response, got "
                + str(response.WhichOneof("response"))
            )
        return response.createPolicy.policyId

    def send_data_room_state_action_request(self, data_room_id: str, action: JSONType) -> JSONType:
        """Send a DCR action request."""
        validated_action = DataRoomActionSchema.model_validate(action)
        request = DcrActionRequest(
            dataRoomId=data_room_id,
            action=validated_action.model_dump_json().encode("utf-8")
        )
        authenticated_request = AuthenticatedRequest(dcrAction=request)
        response = self.send_authenticated_request(authenticated_request)
        if not response.HasField("dcrAction"):
            raise Exception(
                f"Expected `dcrAction` response, got "
                + str(response.WhichOneof("response"))
            )
        return json.loads(response.dcrAction.response)

    def send_retrieve_result_encryption_key_request(self, job_id: str, task_result_hash: str) -> Tuple[str, bytes]:
        """Retrieve the manifest hash and encryption key for a result."""
        request = RetrieveResultEncryptionKeyRequest(
            jobId=job_id,
            taskResultHashHex=task_result_hash
        )
        authenticated_request = AuthenticatedRequest(retrieveResultEncryptionKey=request)
        response = self.send_authenticated_request(authenticated_request)
        if not response.HasField("retrieveResultEncryptionKey"):
            raise Exception(
                f"Expected `retrieveResultEncryptionKey` response, got "
                + str(response.WhichOneof("response"))
            )
        return response.retrieveResultEncryptionKey.manifestHashHex, response.retrieveResultEncryptionKey.encryptionKey

    def send_export_result_as_dataset_request(self, job_id: str, task_result_hash: str, zip_path: Optional[str]) -> Tuple[str, str, str]:
        """Export a result as a dataset."""
        request = ExportResultAsDatasetRequest(
            jobId=job_id,
            taskResultHashHex=task_result_hash,
            raw=RawExport()
        )
        if zip_path is not None:
            request.zip.path = zip_path
        authenticated_request = AuthenticatedRequest(exportResultAsDataset=request)
        response = self.send_authenticated_request(authenticated_request)
        if not response.HasField("exportResultAsDataset"):
            raise Exception(
                f"Expected `exportResultAsDataset` response, got "
                + str(response.WhichOneof("response"))
            )
        return response.exportResultAsDataset.manifestHashHex, response.exportResultAsDataset.secretId, response.exportResultAsDataset.datasetId

    def send_get_verification_key_request(self) -> bytes:
        """Retrieve the verification key for a DCR."""
        request = GetVerificationKeyRequest()
        authenticated_request = AuthenticatedRequest(getVerificationKey=request)
        response = self.send_authenticated_request(authenticated_request)
        if not response.HasField("getVerificationKey"):
            raise Exception(
                f"Expected `getVerificationKey` response, got "
                + str(response.WhichOneof("response"))
            )
        return response.getVerificationKey.verificationKey



from __future__ import annotations

from ...analytics.node_definitions import NodeDefinition
from ...analytics.high_level_node import ComputationNode
from ...session import Session
from typing import Dict
from decentriq_dcr_compiler._schemas.data_science_data_room import (
    SnowflakeConfig,
)
from typing_extensions import Self


class SnowflakeImportConnectorDefinition(NodeDefinition):
    def __init__(
        self,
        name: str,
        warehouse_name: str,
        database_name: str,
        schema_name: str,
        table_name: str,
        stage_name: str,
        credentials_dependency: str,
    ) -> None:
        """
        Initialise a `SnowflakeImportConnectorDefinition`.
        This class is used in order to construct SnowflakeImportConnectors.

        **Parameters**:
        - `name`: Name of the `SnowflakeImportConnectorDefinition`.
        - `warehouse_name`: Name of the warehouse to be used.
        - `database_name`: Name of the database to be used.
        - `schema_name`: Name of the schema to be used.
        - `table_name`: Name of the table where data is imported from.
        - `credentials_dependency`: Name of the credentials node.
        """
        super().__init__(name, id=name)
        self.warehouse_name = warehouse_name
        self.database_name = database_name
        self.schema_name = schema_name
        self.table_name = table_name
        self.stage_name = stage_name
        self.credentials_dependency = credentials_dependency
        self.specification_id = "decentriq.python-ml-worker-32-64"

    def _get_high_level_representation(self) -> Dict[str, str]:
        """
        Retrieve the high level representation of the `SnowflakeImportConnectorDefinition`.
        """
        return {
            "id": self.id,
            "name": self.name,
            "kind": {
                "computation": {
                    "kind": {
                        "importConnector": {
                            "credentialsDependency": self.credentials_dependency,
                            "kind": {
                                "snowflake": {
                                    "databaseName": self.database_name,
                                    "schemaName": self.schema_name,
                                    "stageName": self.stage_name,
                                    "tableName": self.table_name,
                                    "warehouseName": self.warehouse_name,
                                }
                            },
                            "specificationId": self.specification_id,
                        },
                    },
                }
            },
        }

    def build(
        self,
        dcr_id: str,
        node_definition: NodeDefinition,
        client: Client,
        session: Session,
    ) -> SnowflakeImportConnector:
        """
        Construct a SnowflakeImportConnector from the Node Definition.

        **Parameters**:
        - `dcr_id`: ID of the DCR the node is a member of.
        - `node_definition`: Definition of the Import Connector Node.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        """
        return SnowflakeImportConnector(
            name=self.name,
            dcr_id=dcr_id,
            client=client,
            session=session,
            connector_definition=node_definition,
        )

    @property
    def required_workers(self):
        return [self.specification_id]

    @classmethod
    def _from_high_level(
        cls,
        name: str,
        config: SnowflakeConfig,
        credentials_dependency: str,
    ) -> Self:
        """
        Instantiate a `SnowflakeImportConnectorDefinition` from its high level representation.

        **Parameters**:
        - `name`: Name of the `SnowflakeImportConnectorDefinition`.
        - `config`: Pydantic model of the `SnowflakeConfig`.
        - `credentials_dependency`: Name of the credentials dependency node.
        """
        return cls(
            name=name,
            warehouse_name=config.warehouseName,
            database_name=config.databaseName,
            schema_name=config.schemaName,
            table_name=config.tableName,
            stage_name=config.stageName,
            credentials_dependency=credentials_dependency,
        )


class SnowflakeImportConnector(ComputationNode):
    """
    A SnowflakeImportConnector which can import data from Snowflake.
    """

    def __init__(
        self,
        name: str,
        dcr_id: str,
        client: Client,
        session: Session,
        connector_definition: SnowflakeImportConnectorDefinition,
    ) -> None:
        """
        Initialise a `SnowflakeImportConnector`.

        **Parameters**:
        - `name`: Name of the `SnowflakeImportConnector`.
        - `dcr_id`: ID of the DCR the connector is a member of.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        - `connector_definition`: Definition of the Snowflake import connector.
        """
        super().__init__(
            name=name,
            client=client,
            session=session,
            dcr_id=dcr_id,
            id=name,
            definition=connector_definition,
        )
        self.definition = connector_definition

    def _get_computation_id(self) -> str:
        """
        Retrieve the ID of the node corresponding to the connector.
        """
        return f"{self.id}_container"



from .compute_s3_sink_pb2 import (
    FullContent,
    RawObject,
    S3Credentials,
    S3Object,
    S3SinkWorkerConfiguration,
    SingleFile,
    ZipObject,
    DqDspCredentials,
    UserDefinedCredentials,
    DspCredentialsType,
)



from __future__ import annotations

from typing import TYPE_CHECKING

from ...analytics.node_definitions import NodeDefinition
from ...analytics.high_level_node import ComputationNode
from ...session import Session
from decentriq_dcr_compiler._schemas.data_science_data_room import (
    ExportNodeDependency as ExportNodeDependencySchema,
)
from typing_extensions import Self
from ..export_dependency_types import *
from ..export_dependency_types import _get_export_node_dependency_from_high_level


if TYPE_CHECKING:
    from ...client import Client


class AzureBlobStorageExportConnectorDefinition(NodeDefinition):
    def __init__(
        self,
        name: str,
        credentials_dependency: str,
        node_dependency: ExportNodeDependency,
    ) -> None:
        """
        Initialise an `AzureBlobStorageExportConnectorDefinition`.
        This class is used in order to construct an `AzureBlobStorageExportConnector`.

        **Parameters**:
        - `name`: Name of the `AzureBlobStorageExportConnectorDefinition`.
        - `credentials_dependency`: Name of the credentials node.
        - `node_dependency`: The name of the node who's data will be exported to Azure.
                This also defines the type of upload (raw, single file in a zip, entire zip contents).
        """
        super().__init__(name, id=name)
        self.credentials_dependency = credentials_dependency
        self.node_dependency = node_dependency
        self.specification_id = "decentriq.azure-blob-storage-worker"

    def _get_high_level_representation(self) -> dict[str, str]:
        """
        Retrieve the high level representation of the `AzureBlobStorageExportConnectorDefinition`.
        """
        return {
            "id": self.id,
            "name": self.name,
            "kind": {
                "computation": {
                    "kind": {
                        "exportConnector": {
                            "credentialsDependency": {
                                "user": self.credentials_dependency,
                            },
                            "kind": {
                                "azure": (),
                            },
                            "dependency": self.node_dependency.high_level,
                            "specificationId": self.specification_id,
                        },
                    },
                }
            },
        }

    def build(
        self,
        dcr_id: str,
        node_definition: NodeDefinition,
        client: Client,
        session: Session,
    ) -> AzureBlobStorageExportConnector:
        """
        Construct an AzureBlobStorageExportConnector from the definition.

        **Parameters**:
        - `dcr_id`: ID of the DCR the connector is a member of.
        - `node_definition`: Definition of the Azure Blob Storage export connector.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        """
        return AzureBlobStorageExportConnector(
            name=self.name,
            dcr_id=dcr_id,
            client=client,
            session=session,
            connector_definition=node_definition,
        )

    @property
    def required_workers(self):
        return [self.specification_id]

    @classmethod
    def _from_high_level(
        cls,
        name: str,
        credentials_dependency: str,
        node_dependency: ExportNodeDependencySchema,
    ) -> Self:
        """
        Instantiate an `AzureBlobStorageExportConnectorDefinition` from its high level representation.

        **Parameters**:
        - `name`: Name of the import connector.
        - `credentials_dependency`: Name of the node storing the Azure credentials.
        - `node_dependency`: Name of the node whose data should be exported.
        """
        node_dep = _get_export_node_dependency_from_high_level(
            node_dependency, ""
        )
        return cls(
            name=name,
            credentials_dependency=credentials_dependency,
            node_dependency=node_dep,
        )


class AzureBlobStorageExportConnector(ComputationNode):
    """
    An AzureBlobStorageExportConnector which can export data to Azure.

    Note that DCRs containing this type of node cannot be interacted with through the Decentriq UI.
    """

    def __init__(
        self,
        name: str,
        dcr_id: str,
        client: Client,
        session: Session,
        connector_definition: NodeDefinition,
    ) -> None:
        """
        Initialise an `AzureBlobStorageExportConnector`.

        **Parameters**:
        - `name`: Name of the export connector.
        - `dcr_id`: ID of the DCR the connector is a member of.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        - `connector_definition`: Definition of the Azure Blob Storage export connector.
        """
        super().__init__(
            name=name,
            client=client,
            session=session,
            dcr_id=dcr_id,
            id=name,
            definition=connector_definition,
        )
        self.definition = connector_definition

    def _get_computation_id(self) -> str:
        """
        Retrieve the ID of the node corresponding to the computation.
        """
        return self.id



from ..types import DataLabDatasetType, MatchingId, MatchingIdFormat
from .builder import DataLabBuilder
from .data_lab import DataLab

__pdoc__ = {
    "builder": False,
    "data_lab": False,
}

__all__ = [
    "DataLab",
    "DataLabBuilder",
    "DataLabDatasetType",
    "MatchingId",
    "MatchingIdFormat",
]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class EndorsementRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PKIENDORSEMENTREQUEST_FIELD_NUMBER: builtins.int
    DCRSECRETENDORSEMENTREQUEST_FIELD_NUMBER: builtins.int
    @property
    def pkiEndorsementRequest(self) -> global___PkiEndorsementRequest: ...
    @property
    def dcrSecretEndorsementRequest(self) -> global___DcrSecretEndorsementRequest: ...
    def __init__(
        self,
        *,
        pkiEndorsementRequest: global___PkiEndorsementRequest | None = ...,
        dcrSecretEndorsementRequest: global___DcrSecretEndorsementRequest | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["dcrSecretEndorsementRequest", b"dcrSecretEndorsementRequest", "endorsementRequest", b"endorsementRequest", "pkiEndorsementRequest", b"pkiEndorsementRequest"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dcrSecretEndorsementRequest", b"dcrSecretEndorsementRequest", "endorsementRequest", b"endorsementRequest", "pkiEndorsementRequest", b"pkiEndorsementRequest"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["endorsementRequest", b"endorsementRequest"]) -> typing.Literal["pkiEndorsementRequest", "dcrSecretEndorsementRequest"] | None: ...

global___EndorsementRequest = EndorsementRequest

@typing.final
class EndorsementResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PKIENDORSEMENTRESPONSE_FIELD_NUMBER: builtins.int
    DCRSECRETENDORSEMENTRESPONSE_FIELD_NUMBER: builtins.int
    @property
    def pkiEndorsementResponse(self) -> global___PkiEndorsementResponse: ...
    @property
    def dcrSecretEndorsementResponse(self) -> global___DcrSecretEndorsementResponse: ...
    def __init__(
        self,
        *,
        pkiEndorsementResponse: global___PkiEndorsementResponse | None = ...,
        dcrSecretEndorsementResponse: global___DcrSecretEndorsementResponse | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["dcrSecretEndorsementResponse", b"dcrSecretEndorsementResponse", "endorsementResponse", b"endorsementResponse", "pkiEndorsementResponse", b"pkiEndorsementResponse"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dcrSecretEndorsementResponse", b"dcrSecretEndorsementResponse", "endorsementResponse", b"endorsementResponse", "pkiEndorsementResponse", b"pkiEndorsementResponse"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["endorsementResponse", b"endorsementResponse"]) -> typing.Literal["pkiEndorsementResponse", "dcrSecretEndorsementResponse"] | None: ...

global___EndorsementResponse = EndorsementResponse

@typing.final
class PkiEndorsementRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CERTIFICATECHAINPEM_FIELD_NUMBER: builtins.int
    certificateChainPem: builtins.bytes
    def __init__(
        self,
        *,
        certificateChainPem: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["certificateChainPem", b"certificateChainPem"]) -> None: ...

global___PkiEndorsementRequest = PkiEndorsementRequest

@typing.final
class PkiEndorsementResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PKIENDORSEMENT_FIELD_NUMBER: builtins.int
    @property
    def pkiEndorsement(self) -> global___EnclaveEndorsement: ...
    def __init__(
        self,
        *,
        pkiEndorsement: global___EnclaveEndorsement | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["pkiEndorsement", b"pkiEndorsement"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["pkiEndorsement", b"pkiEndorsement"]) -> None: ...

global___PkiEndorsementResponse = PkiEndorsementResponse

@typing.final
class DcrSecretEndorsementRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DCRSECRET_FIELD_NUMBER: builtins.int
    dcrSecret: builtins.str
    def __init__(
        self,
        *,
        dcrSecret: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["dcrSecret", b"dcrSecret"]) -> None: ...

global___DcrSecretEndorsementRequest = DcrSecretEndorsementRequest

@typing.final
class DcrSecretEndorsementResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DCRSECRETENDORSEMENT_FIELD_NUMBER: builtins.int
    DCRSECRETID_FIELD_NUMBER: builtins.int
    dcrSecretId: builtins.bytes
    @property
    def dcrSecretEndorsement(self) -> global___EnclaveEndorsement: ...
    def __init__(
        self,
        *,
        dcrSecretEndorsement: global___EnclaveEndorsement | None = ...,
        dcrSecretId: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["dcrSecretEndorsement", b"dcrSecretEndorsement"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dcrSecretEndorsement", b"dcrSecretEndorsement", "dcrSecretId", b"dcrSecretId"]) -> None: ...

global___DcrSecretEndorsementResponse = DcrSecretEndorsementResponse

@typing.final
class EnclaveEndorsements(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PERSONALPKI_FIELD_NUMBER: builtins.int
    DQPKI_FIELD_NUMBER: builtins.int
    DCRSECRET_FIELD_NUMBER: builtins.int
    @property
    def personalPki(self) -> global___EnclaveEndorsement: ...
    @property
    def dqPki(self) -> global___EnclaveEndorsement: ...
    @property
    def dcrSecret(self) -> global___EnclaveEndorsement: ...
    def __init__(
        self,
        *,
        personalPki: global___EnclaveEndorsement | None = ...,
        dqPki: global___EnclaveEndorsement | None = ...,
        dcrSecret: global___EnclaveEndorsement | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["dcrSecret", b"dcrSecret", "dqPki", b"dqPki", "personalPki", b"personalPki"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dcrSecret", b"dcrSecret", "dqPki", b"dqPki", "personalPki", b"personalPki"]) -> None: ...

global___EnclaveEndorsements = EnclaveEndorsements

@typing.final
class EnclaveEndorsement(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENDORSEMENTCERTIFICATEDER_FIELD_NUMBER: builtins.int
    endorsementCertificateDer: builtins.bytes
    def __init__(
        self,
        *,
        endorsementCertificateDer: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["endorsementCertificateDer", b"endorsementCertificateDer"]) -> None: ...

global___EnclaveEndorsement = EnclaveEndorsement

@typing.final
class PkiClaim(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ROOTCERTIFICATEDER_FIELD_NUMBER: builtins.int
    rootCertificateDer: builtins.bytes
    def __init__(
        self,
        *,
        rootCertificateDer: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["rootCertificateDer", b"rootCertificateDer"]) -> None: ...

global___PkiClaim = PkiClaim

@typing.final
class DcrSecretClaim(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DCRSECRETID_FIELD_NUMBER: builtins.int
    dcrSecretId: builtins.bytes
    def __init__(
        self,
        *,
        dcrSecretId: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["dcrSecretId", b"dcrSecretId"]) -> None: ...

global___DcrSecretClaim = DcrSecretClaim



"""
WARNING: This file is copied from trusted/python-libs/dq-media-dcr-api/dq_media_dcr_api.
DO NOT EDIT THIS FILE DIRECTLY - make changes in the source repository instead and run the copy_init.py script.
"""

"""
This module defines the DCR API models for the types used by dq-media-dcr.
To ensure backwards compatibility, new fields should be added as optional with a default value of None.
Marking a field as optional is not enough, a default **must** be provided. Failure to do so will result in a validation error
on the client side.
"""

import warnings
from enum import Enum
from typing import List, Literal, Optional, Union, Dict, Any, Sequence
from datetime import datetime
from pydantic import BaseModel, Field, conint

# Suppress Pydantic shadowing warnings
warnings.filterwarnings("ignore", message="Field name.*shadows an attribute.*BaseModel")


class AudienceKind(str, Enum):
    SEED = "SEED"
    CUSTOM = "CUSTOM"


class SeedAudience(BaseModel):
    kind: AudienceKind
    audienceType: str
    audienceSize: Optional[int] = None
    createdAt: datetime
    isEmpty: bool


class AudienceReference1(BaseModel):
    kind: Literal["BASE"]


class AudienceReference2(BaseModel):
    id: str
    kind: Literal["CUSTOM"]


class AudienceReference3(BaseModel):
    audienceType: str
    kind: Literal["SEED"]


class SeedAudienceReference(BaseModel):
    audienceType: str


AudienceReference = Union[AudienceReference1, AudienceReference2, AudienceReference3]


class CustomAudienceDefinition1(BaseModel):
    excludeSeedAudience: Optional[bool] = None
    kind: Literal["LOOKALIKE"]
    reach: int
    sourceRef: AudienceReference


CombineOperator = Literal["INTERSECT", "UNION", "DIFF"]
FilterOperator = Literal[
    "CONTAINS_ANY_OF", "CONTAINS_NONE_OF", "CONTAINS_ALL_OF", "EMPTY", "NOT_EMPTY"
]


class AudienceFilter(BaseModel):
    attribute: str
    operator: FilterOperator
    values: Optional[Sequence[str]] = None


BooleanOp = Literal["AND", "OR"]


class AudienceFilters(BaseModel):
    booleanOp: BooleanOp
    filters: Sequence[AudienceFilter]


class AudienceCombinator(BaseModel):
    filters: Optional[AudienceFilters] = None
    operator: CombineOperator
    sourceRef: AudienceReference


class CustomAudienceDefinition2(BaseModel):
    combine: Optional[Sequence[AudienceCombinator]] = None
    filters: Optional[AudienceFilters] = None
    kind: Literal["RULE_BASED"]
    sourceRef: AudienceReference


class CustomAudienceDefinition3(BaseModel):
    kind: Literal["REMARKETING"]
    seedAudienceRef: SeedAudienceReference


class StalenessWarnings(BaseModel):
    missingCustomAudiences: Optional[Sequence[str]] = []
    missingFilterAttributes: Optional[Sequence[str]] = []
    missingFilterValues: Optional[Sequence[str]] = []
    missingSeedAudiences: Optional[Sequence[str]] = []


class CustomAudience(BaseModel):
    createdAt: str
    createdBy: str
    definition: Union[CustomAudienceDefinition1, CustomAudienceDefinition2, CustomAudienceDefinition3]
    doesAudienceDependOnLookalikeAudience: Optional[bool] = None
    id: str
    kind: AudienceKind = AudienceKind.CUSTOM
    name: str
    sharedWith: Optional[Sequence[str]] = None
    stalenessWarnings: Optional[StalenessWarnings] = None


class GetCustomAudiencesResult(BaseModel):
    audiences: List[CustomAudience]


class GetSeedAudiencesResult(BaseModel):
    audiences: List[SeedAudience]


class ValueType(str, Enum):
    NUMERIC_RANGE = "NUMERIC_RANGE"
    CATEGORICAL = "CATEGORICAL"


class FilterType(str, Enum):
    CONTAINS_NONE_OF = "CONTAINS_NONE_OF"
    CONTAINS_ANY_OF = "CONTAINS_ANY_OF"
    CONTAINS_ALL_OF = "CONTAINS_ALL_OF"
    EMPTY = "EMPTY"
    NOT_EMPTY = "NOT_EMPTY"


class AttributeDefinition(BaseModel):
    values: List[Union[str, int, float]]
    value_type: ValueType
    filters: List[FilterType]


class DataAttributesResult(BaseModel):
    attributes: Dict[str, AttributeDefinition]


class AggregationColumn(BaseModel):
    column: str
    possible_values: List[str]


class Aggregation(BaseModel):
    id: str
    aggregation: List[AggregationColumn]
    audience_type: str
    columns: List[str]
    rows: List[List[Any]]


class ComputeInsightsAudience(BaseModel):
    audience_type: str
    aggregations: List[Aggregation]


class ComputeInsightsResult(BaseModel):
    total_publisher: Optional[int] = (
        None  # Optional because it depends on hide_absolute_values_from_insights
    )
    audiences: List[ComputeInsightsAudience]


class OverlapStatistic(BaseModel):
    audience_type: str
    advertiser_size: Optional[int] = None  # Optional because it might be hidden
    advertiser_is_empty: bool
    overlap_size: Optional[int] = None  # Optional because it might be hidden
    match_rate: Optional[float] = None  # Optional because it might be hidden


class OverlapStatisticsResult(BaseModel):
    overlap_statistics: List[OverlapStatistic]


class RocCurvePoint(BaseModel):
    """Represents a single point on the ROC curve"""

    tpr: float
    fpr: float


class FilteredAudienceQuality(BaseModel):
    """Quality metrics for a filtered audience"""

    roc_curve: Optional[RocCurvePoint] = None


class FilteredAudience(BaseModel):
    """Represents a filtered audience at a specific reach percentage"""

    reach: int
    size: int
    quality: FilteredAudienceQuality


class RocCurveData(BaseModel):
    """Complete ROC curve data including AUC and test set information"""

    auc: float
    fpr: List[float]
    tpr: List[float]
    reach: List[float]
    test_set_size: int
    test_set_positive_examples_size: int


class QualityMetrics(BaseModel):
    """Overall quality metrics including complete ROC curve data"""

    roc_curve: RocCurveData


class LookalikeAudienceStatisticsResult(BaseModel):
    """Root model representing the complete lookalike audience statistics"""

    addressable_audience_size: int
    seed_audience_size: int
    filtered_audiences: List[FilteredAudience]
    quality: QualityMetrics


class CellLocation(BaseModel):
    column: int = Field(ge=0)
    row: int = Field(ge=0)


class CellValidationErrorCode(str, Enum):
    EMPTY_CELL = "EMPTY_CELL"
    INTERNAL_VALIDATION_ERROR = "INTERNAL_VALIDATION_ERROR"
    INVALID_NUMBER_FORMAT = "INVALID_NUMBER_FORMAT"
    INVALID_INTEGER_FORMAT = "INVALID_INTEGER_FORMAT"
    INVALID_EMAIL_FORMAT = "INVALID_EMAIL_FORMAT"
    INVALID_ISO_DATE_FORMAT = "INVALID_ISO_DATE_FORMAT"
    INVALID_PHONE_E164_NUMBER = "INVALID_PHONE_E164_NUMBER"
    INVALID_FLOAT_FORMAT = "INVALID_FLOAT_FORMAT"
    INVALID_SHA256_HEX_FORMAT = "INVALID_SHA256_HEX_FORMAT"
    INVALID_NUMERIC_RANGE = "INVALID_NUMERIC_RANGE"


class CellValidationError(BaseModel):
    code: CellValidationErrorCode
    location: CellLocation
    message: str


class DuplicationValidationErrorCode(str, Enum):
    DUPLICATE_VALUES = "DUPLICATE_VALUES"


class MultiColumnLocation(BaseModel):
    columns: Sequence[conint(ge=0)]
    row: int = Field(ge=0)


class DuplicationValidationError(BaseModel):
    code: DuplicationValidationErrorCode
    location: MultiColumnLocation


class SchemaValidationErrorCode(str, Enum):
    INTERNAL_VALIDATION_ERROR = "INTERNAL_VALIDATION_ERROR"
    DUPLICATE_VALUES = "DUPLICATE_VALUES"
    NOT_ENOUGH_COLUMNS_PROVIDED = "NOT_ENOUGH_COLUMNS_PROVIDED"
    TOO_MANY_COLUMNS_PROVIDED = "TOO_MANY_COLUMNS_PROVIDED"
    EMPTY_FILE = "EMPTY_FILE"


class SchemaValidationError(BaseModel):
    code: SchemaValidationErrorCode
    message: str
    row: int = Field(ge=0)


class TableValidationErrorCode(str, Enum):
    EMPTY_FILE = "EMPTY_FILE"
    INVALID_NUMBER_OF_ROWS = "INVALID_NUMBER_OF_ROWS"


class TableValidationError(BaseModel):
    code: TableValidationErrorCode
    message: str


class ValidatedColumnV0(BaseModel):
    column: int = Field(ge=0)
    numErrorsTotal: int = Field(ge=0)
    recordedErrors: Sequence[CellValidationError]


class ValidatedSchemaV0(BaseModel):
    numErrorsTotal: int = Field(ge=0)
    recordedErrors: Sequence[SchemaValidationError]


class ValidatedTableV0(BaseModel):
    numErrorsTotal: int = Field(ge=0)
    recordedErrors: Sequence[TableValidationError]


class ValidatedUniqueness(BaseModel):
    numErrorsTotal: int = Field(ge=0)
    recordedErrors: Sequence[DuplicationValidationError]


class ValidationOutcome(str, Enum):
    PASSED = "PASSED"
    FAILED = "FAILED"


class DropInvalidRows(BaseModel):
    numInvalidRowsDropped: int = Field(ge=0)


class ValidationReportV0(BaseModel):
    columns: Sequence[ValidatedColumnV0]
    dropInvalidRows: Optional[DropInvalidRows] = None
    numInvalidRowsTotal: int = Field(ge=0)
    numRowsTotal: int = Field(default=0, ge=0)
    originalFileSizeBytes: Optional[int] = Field(None, ge=0)
    outcome: ValidationOutcome
    schema: ValidatedSchemaV0  # Shadows BaseModel.schema
    table: ValidatedTableV0
    uniqueness: Optional[ValidatedUniqueness] = None


class ValidationReport(BaseModel):
    report: ValidationReportV0
    version: str = Field(..., enum=["v0"])



import json
from .api import ValidationReport


class ValidationReports:
    """
    A class to represent the validation reports for a DCR.
    """

    def __init__(
        self,
        matching: ValidationReport,
        segments: ValidationReport,
        demographics: ValidationReport,
    ):
        """
        Initialize a ValidationReports instance.

        **Parameters**:
        - `matching`: The matching validation report
        - `segments`: The segments validation report
        - `demographics`: The demographics validation report
        """
        self.formatted_reports = {
            "matching": matching.model_dump(),
            "segments": segments.model_dump(),
            "demographics": demographics.model_dump(),
        }

    def __str__(self):
        """
        Return a JSON string representation of the validation reports.
        """
        return f"{json.dumps(self.formatted_reports, indent=2)}"

    # Added to mimic the behaviour of Pydantic types to provide a
    # consistent interface for the user.
    def model_dump_json(self):
        """
        Return a JSON string representation of the validation reports.
        """
        return json.dumps(self.formatted_reports, indent=2)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class IntegrityStoreRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CAS_FIELD_NUMBER: builtins.int
    @property
    def cas(self) -> global___BatchCompareAndSwapRequest: ...
    def __init__(
        self,
        *,
        cas: global___BatchCompareAndSwapRequest | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["cas", b"cas", "request", b"request"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["cas", b"cas", "request", b"request"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["request", b"request"]) -> typing.Literal["cas"] | None: ...

global___IntegrityStoreRequest = IntegrityStoreRequest

@typing.final
class IntegrityStoreResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CAS_FIELD_NUMBER: builtins.int
    @property
    def cas(self) -> global___BatchCompareAndSwapResponse: ...
    def __init__(
        self,
        *,
        cas: global___BatchCompareAndSwapResponse | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["cas", b"cas", "response", b"response"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["cas", b"cas", "response", b"response"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["response", b"response"]) -> typing.Literal["cas"] | None: ...

global___IntegrityStoreResponse = IntegrityStoreResponse

@typing.final
class BatchCompareAndSwapRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REQUESTS_FIELD_NUMBER: builtins.int
    @property
    def requests(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___CompareAndSwapRequest]: ...
    def __init__(
        self,
        *,
        requests: collections.abc.Iterable[global___CompareAndSwapRequest] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["requests", b"requests"]) -> None: ...

global___BatchCompareAndSwapRequest = BatchCompareAndSwapRequest

@typing.final
class BatchCompareAndSwapResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RESPONSES_FIELD_NUMBER: builtins.int
    @property
    def responses(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___CompareAndSwapResponse]: ...
    def __init__(
        self,
        *,
        responses: collections.abc.Iterable[global___CompareAndSwapResponse] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["responses", b"responses"]) -> None: ...

global___BatchCompareAndSwapResponse = BatchCompareAndSwapResponse

@typing.final
class CompareAndSwapRequest(google.protobuf.message.Message):
    """Try to store an entry under the given id.
    If 'expectedCasIndex' is null, the request will only succeed if there is no existing entry stored under that id.
    Otherwise, the store will check that the existing entry is at `expectedCasIndex`, and if it is, it is swapped with 'newValue'.
    If 'newValue' is null then the entry is tombstoned instead. Note that both initial insertion and deletion are atomic as
    well, i.e. two parallel insertions or deletions will have only one succeed.

    Creating an entry -> expectedCasIndex == null, newEntry != null
    Getting an entry -> expectedCasIndex == null, newEntry == null
    Deleting an entry -> expectedCasIndex != null, newEntry == null
    Updating an entry -> expectedCasIndex != null, newEntry != null
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    CONTENT_TYPE_FIELD_NUMBER: builtins.int
    EXPECTEDCASINDEX_FIELD_NUMBER: builtins.int
    NEWVALUE_FIELD_NUMBER: builtins.int
    id: builtins.str
    content_type: builtins.str
    expectedCasIndex: builtins.int
    @property
    def newValue(self) -> global___IntegrityStoreValue: ...
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        content_type: builtins.str = ...,
        expectedCasIndex: builtins.int | None = ...,
        newValue: global___IntegrityStoreValue | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_expectedCasIndex", b"_expectedCasIndex", "_newValue", b"_newValue", "expectedCasIndex", b"expectedCasIndex", "newValue", b"newValue"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_expectedCasIndex", b"_expectedCasIndex", "_newValue", b"_newValue", "content_type", b"content_type", "expectedCasIndex", b"expectedCasIndex", "id", b"id", "newValue", b"newValue"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_expectedCasIndex", b"_expectedCasIndex"]) -> typing.Literal["expectedCasIndex"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_newValue", b"_newValue"]) -> typing.Literal["newValue"] | None: ...

global___CompareAndSwapRequest = CompareAndSwapRequest

@typing.final
class CompareAndSwapResponse(google.protobuf.message.Message):
    """Response of a CAS operation.
    'swapped' indicates whether the swapping operation was successful. Regardless of success the previous value is
    returned, null meaning that there was no value associated with the id.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SWAPPED_FIELD_NUMBER: builtins.int
    PREVIOUSCASINDEX_FIELD_NUMBER: builtins.int
    PREVIOUSVALUE_FIELD_NUMBER: builtins.int
    swapped: builtins.bool
    previousCasIndex: builtins.int
    @property
    def previousValue(self) -> global___IntegrityStoreValue: ...
    def __init__(
        self,
        *,
        swapped: builtins.bool = ...,
        previousCasIndex: builtins.int | None = ...,
        previousValue: global___IntegrityStoreValue | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_previousCasIndex", b"_previousCasIndex", "_previousValue", b"_previousValue", "previousCasIndex", b"previousCasIndex", "previousValue", b"previousValue"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_previousCasIndex", b"_previousCasIndex", "_previousValue", b"_previousValue", "previousCasIndex", b"previousCasIndex", "previousValue", b"previousValue", "swapped", b"swapped"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_previousCasIndex", b"_previousCasIndex"]) -> typing.Literal["previousCasIndex"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_previousValue", b"_previousValue"]) -> typing.Literal["previousValue"] | None: ...

global___CompareAndSwapResponse = CompareAndSwapResponse

@typing.final
class IntegrityStoreValue(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VERSION_FIELD_NUMBER: builtins.int
    PLAINTEXTCONTENT_FIELD_NUMBER: builtins.int
    PLAINTEXTSIGNATURE_FIELD_NUMBER: builtins.int
    ENCRYPTEDCONTENT_FIELD_NUMBER: builtins.int
    version: builtins.int
    """/ Specifies how to interpret the "content" and "signature" fields
    / (e.g. what algorithm was used when creating the signature).
    """
    plaintextContent: builtins.str
    """/ Generic stringified payload"""
    plaintextSignature: builtins.bytes
    """/ Signature over (content, contentType, version, id)"""
    encryptedContent: builtins.bytes
    """/ Encrypted payload"""
    def __init__(
        self,
        *,
        version: builtins.int = ...,
        plaintextContent: builtins.str | None = ...,
        plaintextSignature: builtins.bytes | None = ...,
        encryptedContent: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_encryptedContent", b"_encryptedContent", "_plaintextContent", b"_plaintextContent", "_plaintextSignature", b"_plaintextSignature", "encryptedContent", b"encryptedContent", "plaintextContent", b"plaintextContent", "plaintextSignature", b"plaintextSignature"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_encryptedContent", b"_encryptedContent", "_plaintextContent", b"_plaintextContent", "_plaintextSignature", b"_plaintextSignature", "encryptedContent", b"encryptedContent", "plaintextContent", b"plaintextContent", "plaintextSignature", b"plaintextSignature", "version", b"version"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_encryptedContent", b"_encryptedContent"]) -> typing.Literal["encryptedContent"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_plaintextContent", b"_plaintextContent"]) -> typing.Literal["plaintextContent"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_plaintextSignature", b"_plaintextSignature"]) -> typing.Literal["plaintextSignature"] | None: ...

global___IntegrityStoreValue = IntegrityStoreValue



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class SecretStoreEntry(google.protobuf.message.Message):
    """An entry in the secret store"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTENT_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    content: builtins.bytes
    state: builtins.bytes
    def __init__(
        self,
        *,
        content: builtins.bytes = ...,
        state: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["content", b"content", "state", b"state"]) -> None: ...

global___SecretStoreEntry = SecretStoreEntry

@typing.final
class CreateSecretRequest(google.protobuf.message.Message):
    """/ Add a new secret to the store"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SECRET_FIELD_NUMBER: builtins.int
    @property
    def secret(self) -> global___SecretStoreEntry: ...
    def __init__(
        self,
        *,
        secret: global___SecretStoreEntry | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["secret", b"secret"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["secret", b"secret"]) -> None: ...

global___CreateSecretRequest = CreateSecretRequest

@typing.final
class CreateSecretResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    id: builtins.str
    def __init__(
        self,
        *,
        id: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["id", b"id"]) -> None: ...

global___CreateSecretResponse = CreateSecretResponse

@typing.final
class GetSecretRequest(google.protobuf.message.Message):
    """/ Get a single secret from the secret store"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    id: builtins.str
    version: builtins.int
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        version: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["id", b"id", "version", b"version"]) -> None: ...

global___GetSecretRequest = GetSecretRequest

@typing.final
class GetSecretResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SECRET_FIELD_NUMBER: builtins.int
    CASINDEX_FIELD_NUMBER: builtins.int
    casIndex: builtins.int
    @property
    def secret(self) -> global___SecretStoreEntry: ...
    def __init__(
        self,
        *,
        secret: global___SecretStoreEntry | None = ...,
        casIndex: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["secret", b"secret"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["casIndex", b"casIndex", "secret", b"secret"]) -> None: ...

global___GetSecretResponse = GetSecretResponse

@typing.final
class RemoveSecretRequest(google.protobuf.message.Message):
    """/ Remove a particular secret"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    EXPECTEDCASINDEX_FIELD_NUMBER: builtins.int
    id: builtins.str
    expectedCasIndex: builtins.int
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        expectedCasIndex: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["expectedCasIndex", b"expectedCasIndex", "id", b"id"]) -> None: ...

global___RemoveSecretRequest = RemoveSecretRequest

@typing.final
class RemoveSecretResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REMOVED_FIELD_NUMBER: builtins.int
    removed: builtins.bool
    """/ Removal can fail because the expected cas index doesn't match"""
    def __init__(
        self,
        *,
        removed: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["removed", b"removed"]) -> None: ...

global___RemoveSecretResponse = RemoveSecretResponse

@typing.final
class UpdateSecretAclRequest(google.protobuf.message.Message):
    """/ Remove a particular secret"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    NEWACL_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    EXPECTEDCASINDEX_FIELD_NUMBER: builtins.int
    id: builtins.str
    newAcl: builtins.bytes
    version: builtins.int
    expectedCasIndex: builtins.int
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        newAcl: builtins.bytes = ...,
        version: builtins.int = ...,
        expectedCasIndex: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["expectedCasIndex", b"expectedCasIndex", "id", b"id", "newAcl", b"newAcl", "version", b"version"]) -> None: ...

global___UpdateSecretAclRequest = UpdateSecretAclRequest

@typing.final
class UpdateSecretAclResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    UPDATED_FIELD_NUMBER: builtins.int
    updated: builtins.bool
    """/ Update can fail because the expected cas index doesn't match"""
    def __init__(
        self,
        *,
        updated: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["updated", b"updated"]) -> None: ...

global___UpdateSecretAclResponse = UpdateSecretAclResponse

@typing.final
class SecretStoreRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATESECRET_FIELD_NUMBER: builtins.int
    GETSECRET_FIELD_NUMBER: builtins.int
    REMOVESECRET_FIELD_NUMBER: builtins.int
    UPDATESECRETACL_FIELD_NUMBER: builtins.int
    @property
    def createSecret(self) -> global___CreateSecretRequest: ...
    @property
    def getSecret(self) -> global___GetSecretRequest: ...
    @property
    def removeSecret(self) -> global___RemoveSecretRequest: ...
    @property
    def updateSecretAcl(self) -> global___UpdateSecretAclRequest: ...
    def __init__(
        self,
        *,
        createSecret: global___CreateSecretRequest | None = ...,
        getSecret: global___GetSecretRequest | None = ...,
        removeSecret: global___RemoveSecretRequest | None = ...,
        updateSecretAcl: global___UpdateSecretAclRequest | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["createSecret", b"createSecret", "getSecret", b"getSecret", "removeSecret", b"removeSecret", "request", b"request", "updateSecretAcl", b"updateSecretAcl"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["createSecret", b"createSecret", "getSecret", b"getSecret", "removeSecret", b"removeSecret", "request", b"request", "updateSecretAcl", b"updateSecretAcl"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["request", b"request"]) -> typing.Literal["createSecret", "getSecret", "removeSecret", "updateSecretAcl"] | None: ...

global___SecretStoreRequest = SecretStoreRequest

@typing.final
class SecretStoreResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATESECRET_FIELD_NUMBER: builtins.int
    GETSECRET_FIELD_NUMBER: builtins.int
    REMOVESECRET_FIELD_NUMBER: builtins.int
    UPDATESECRETACL_FIELD_NUMBER: builtins.int
    @property
    def createSecret(self) -> global___CreateSecretResponse: ...
    @property
    def getSecret(self) -> global___GetSecretResponse: ...
    @property
    def removeSecret(self) -> global___RemoveSecretResponse: ...
    @property
    def updateSecretAcl(self) -> global___UpdateSecretAclResponse: ...
    def __init__(
        self,
        *,
        createSecret: global___CreateSecretResponse | None = ...,
        getSecret: global___GetSecretResponse | None = ...,
        removeSecret: global___RemoveSecretResponse | None = ...,
        updateSecretAcl: global___UpdateSecretAclResponse | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["createSecret", b"createSecret", "getSecret", b"getSecret", "removeSecret", b"removeSecret", "response", b"response", "updateSecretAcl", b"updateSecretAcl"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["createSecret", b"createSecret", "getSecret", b"getSecret", "removeSecret", b"removeSecret", "response", b"response", "updateSecretAcl", b"updateSecretAcl"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["response", b"response"]) -> typing.Literal["createSecret", "getSecret", "removeSecret", "updateSecretAcl"] | None: ...

global___SecretStoreResponse = SecretStoreResponse



import base64
import io
import json
import uuid
import zipfile
from pathlib import Path
from typing import Dict, Mapping, Optional, Text, Tuple
from uuid import uuid4

from decentriq_dcr_compiler import compiler
from decentriq_dcr_compiler._schemas.create_data_lab import (
    EnclaveSpecification as HlEnclaveSpecification,
)
from decentriq_dcr_compiler.schemas import (
    CreateDataLab,
    CreateDataLab16,
    CreateDataLabComputeV15,
)

from ..client import Client, SecretStoreOptions
from ..helpers import (
    create_session_from_driver_spec,
    get_latest_enclave_specs_as_dictionary,
)
from ..proto import (
    CreateDcrPurpose,
    DataRoom,
    parse_length_delimited,
    serialize_length_delimited,
)
from ..proto.length_delimited import parse_length_delimited, serialize_length_delimited
from ..session import LATEST_GCG_PROTOCOL_VERSION, Session
from ..storage import Key
from ..types import (
    MATCHING_ID_INTERNAL_LOOKUP,
    DataLabDatasetType,
    DataLabDefinition,
    DryRunOptions,
    EnclaveSpecification,
    JobId,
    MatchingId,
)

class Dataset:
    def __init__(self, manifest_hash: str, key: Key):
        self.manifest_hash = manifest_hash
        self.key = key


class DataLabConfig:
    def __init__(
        self,
        name: str,
        has_demographics: bool,
        has_embeddings: bool,
        num_embeddings: int,
        has_segments: bool,
        matching_id: MatchingId,
        force_spark_validation: bool = False,
        drop_invalid_rows: bool = True
    ):
        self.name = name
        self.has_demographics = has_demographics
        self.has_embeddings = has_embeddings
        self.num_embeddings = num_embeddings
        self.has_segments = has_segments
        self.matching_id = matching_id
        self._force_spark_validation = force_spark_validation
        self.drop_invalid_rows = drop_invalid_rows


class ExistingDataLab:
    def __init__(
        self,
        data_lab_definition: DataLabDefinition,
    ):
        self.id = data_lab_definition["id"]
        self.high_level_representation = data_lab_definition[
            "highLevelRepresentationAsString"
        ]
        self.match_dataset = data_lab_definition["usersDataset"]
        self.segments_dataset = data_lab_definition["segmentsDataset"]
        self.demographics_dataset = data_lab_definition["demographicsDataset"]
        self.embeddings_dataset = data_lab_definition["embeddingsDataset"]


class DataLab:
    def __init__(
        self,
        client: Client,
        cfg: DataLabConfig,
        existing_data_lab: Optional[ExistingDataLab] = None,
    ):
        self.client = client
        enclave_specs = get_latest_enclave_specs_as_dictionary(self.client)
        auth, _ = client.create_auth_using_decentriq_pki(enclave_specs)
        self.session = client.create_session(auth, enclave_specs)
        self.cfg = cfg
        self.dcr_id = ""
        self.datasets: Dict[DataLabDatasetType, Dataset] = dict()

        if existing_data_lab:
            # Create a DataLab using the existing configuration.
            self.data_lab_id = existing_data_lab.id
            self.hl_data_lab = compiler.DataLab.parse_obj(
                json.loads(existing_data_lab.high_level_representation)
            )
            self._populate_datasets_cache_from_existing(existing_data_lab)
        else:
            # Create a new DataLab.
            self.data_lab_id = f"DataLab-{str(uuid.uuid4())}"
            (
                matching_id_format,
                matching_id_hashing_algorithm,
            ) = MATCHING_ID_INTERNAL_LOOKUP[self.cfg.matching_id]
            create_data_lab = CreateDataLab(
                root=CreateDataLab16(
                    v15=CreateDataLabComputeV15(
                        authenticationRootCertificatePem=self.client.decentriq_ca_root_certificate.decode(),
                        driverEnclaveSpecification=HlEnclaveSpecification(
                            attestationProtoBase64="",
                            id="",
                            workerProtocol=0,
                        ),
                        hasDemographics=self.cfg.has_demographics,
                        hasEmbeddings=self.cfg.has_embeddings,
                        hasSegments=self.cfg.has_segments,
                        id=self.data_lab_id,
                        matchingIdFormat=matching_id_format.value,
                        matchingIdHashingAlgorithm=(
                            None
                            if matching_id_hashing_algorithm is None
                            else matching_id_hashing_algorithm.value
                        ),
                        name=self.cfg.name,
                        numEmbeddings=self.cfg.num_embeddings,
                        publisherEmail=self.client.user_email,
                        pythonEnclaveSpecification=HlEnclaveSpecification(
                            attestationProtoBase64="",
                            id="",
                            workerProtocol=0,
                        ),
                        forceSparkValidation=self.cfg._force_spark_validation,
                        dropInvalidRows=self.cfg.drop_invalid_rows,
                    ),
                )
            )
            self.hl_data_lab = compiler.create_data_lab(create_data_lab)
            self.data_lab_id = self.client._publish_data_lab_from_existing(
                json.loads(self.hl_data_lab.json())
            )

    def _populate_datasets_cache_from_existing(
        self, existing_data_lab: ExistingDataLab
    ):
        match_dataset = existing_data_lab.match_dataset
        if match_dataset is not None:
            manifest_hash = match_dataset["manifestHash"]
            key = self.client.get_dataset_key(manifest_hash)
            self.datasets[DataLabDatasetType.MATCH] = Dataset(
                manifest_hash, key
            )

        segments_dataset = existing_data_lab.segments_dataset
        if segments_dataset is not None:
            manifest_hash = segments_dataset["manifestHash"]
            key = self.client.get_dataset_key(manifest_hash)
            self.datasets[DataLabDatasetType.SEGMENTS] = Dataset(
                manifest_hash, key
            )

        demographics_dataset = existing_data_lab.demographics_dataset
        if self.cfg.has_demographics and demographics_dataset is not None:
            manifest_hash = demographics_dataset["manifestHash"]
            key = self.client.get_dataset_key(manifest_hash)
            self.datasets[DataLabDatasetType.DEMOGRAPHICS] = Dataset(
                manifest_hash, key
            )

        embeddings_dataset = existing_data_lab.embeddings_dataset
        if self.cfg.has_embeddings and embeddings_dataset is not None:
            manifest_hash = embeddings_dataset["manifestHash"]
            key = self.client.get_dataset_key(manifest_hash)
            self.datasets[DataLabDatasetType.EMBEDDINGS] = Dataset(
                manifest_hash, key
            )

    def _get_data_lab_enclave_specs(
        self,
        enclave_specs: Dict[str, EnclaveSpecification],
    ) -> Tuple[HlEnclaveSpecification, HlEnclaveSpecification]:
        driver_spec = None
        python_spec = None
        for spec_id, spec in enclave_specs.items():
            spec_payload = {
                "attestationProtoBase64": base64.b64encode(
                    serialize_length_delimited(spec["proto"])
                ).decode(),
                "id": spec_id,
                "workerProtocol": spec["workerProtocols"][0],
            }
            if "decentriq.driver" in spec_id:
                spec["clientProtocols"] = [LATEST_GCG_PROTOCOL_VERSION]
                driver_spec = HlEnclaveSpecification.parse_obj(spec_payload)
            elif "decentriq.python-ml-worker" in spec_id:
                spec["clientProtocols"] = [LATEST_GCG_PROTOCOL_VERSION]
                python_spec = HlEnclaveSpecification.parse_obj(spec_payload)
        if driver_spec is None:
            raise Exception("No driver enclave spec found for the datalab")
        if python_spec is None:
            raise Exception("No python-ml-worker enclave spec found for the datalab")
        return (driver_spec, python_spec)

    def provision_local_datasets(
        self,
        key: Key,
        matching_data_path: str,
        segments_data_path: Optional[str] = None,
        demographics_data_path: Optional[str] = None,
        embeddings_data_path: Optional[str] = None,
        *,
        secret_store_options: Optional[SecretStoreOptions] = None,
    ):
        """
        Upload local datasets and provision to the DataLab.

        **Parameters**
        - `key`: The key used to encrypt the dataset.
        - `match`: The file path to the "match" dataset.
        - `segments`: The file path to the "segments" dataset.
        - `demographics`: The file path to the "demographics" dataset.
        - `embeddings`: The file path to the "embeddings" dataset.
        """
        if matching_data_path is not None:
            dataset_name = Path(matching_data_path).stem
            dataset_id = self._upload_dataset(
                matching_data_path, dataset_name, key, secret_store_options
            )
            self.provision_dataset(dataset_id, key, DataLabDatasetType.MATCH)
        if segments_data_path is not None:
            dataset_name = Path(segments_data_path).stem
            dataset_id = self._upload_dataset(
                segments_data_path, dataset_name, key, secret_store_options
            )
            self.provision_dataset(dataset_id, key, DataLabDatasetType.SEGMENTS)
        if embeddings_data_path is not None:
            dataset_name = Path(embeddings_data_path).stem
            self._check_dataset_type_permitted(
                dataset_type=DataLabDatasetType.EMBEDDINGS
            )
            dataset_id = self._upload_dataset(
                embeddings_data_path, dataset_name, key, secret_store_options
            )
            self.provision_dataset(dataset_id, key, DataLabDatasetType.EMBEDDINGS)
        if demographics_data_path is not None:
            dataset_name = Path(demographics_data_path).stem
            self._check_dataset_type_permitted(
                dataset_type=DataLabDatasetType.DEMOGRAPHICS
            )
            dataset_id = self._upload_dataset(
                demographics_data_path, dataset_name, key, secret_store_options
            )
            self.provision_dataset(dataset_id, key, DataLabDatasetType.DEMOGRAPHICS)

    def _upload_dataset(
        self, file_path: str, name: str, key: Key, secret_store_options: Optional[SecretStoreOptions]
    ):
        with open(file_path, "rb") as file:
            dataset_id = self.client.upload_dataset(file, key, name, secret_store_options=secret_store_options)
            return dataset_id

    def provision_dataset(
        self, manifest_hash: str, key: Key, dataset_type: DataLabDatasetType
    ):
        self._check_dataset_type_permitted(dataset_type=dataset_type)

        if dataset_type == DataLabDatasetType.EMBEDDINGS:
            self.client._set_datalab_embeddings_dataset(self.data_lab_id, manifest_hash)
        elif dataset_type == DataLabDatasetType.DEMOGRAPHICS:
            self.client._set_datalab_demographics_dataset(
                self.data_lab_id, manifest_hash
            )
        elif dataset_type == DataLabDatasetType.MATCH:
            self.client._set_datalab_matching_dataset(self.data_lab_id, manifest_hash)
        elif dataset_type == DataLabDatasetType.SEGMENTS:
            self.client._set_datalab_segments_dataset(self.data_lab_id, manifest_hash)
        else:
            raise Exception(f"Unknown dataset type {dataset_type}")

        # Store the provisioned datasets in the local cache.
        self.datasets[dataset_type] = Dataset(manifest_hash, key)

    def deprovision_dataset(self, dataset_type: DataLabDatasetType):
        self._check_dataset_type_permitted(dataset_type=dataset_type)

        # Sending "None" as the manifest hash causes the dataset to be deprovisioned.
        if dataset_type == DataLabDatasetType.EMBEDDINGS:
            self.client._set_datalab_embeddings_dataset(self.data_lab_id, None)
        elif dataset_type == DataLabDatasetType.DEMOGRAPHICS:
            self.client._set_datalab_demographics_dataset(self.data_lab_id, None)
        elif dataset_type == DataLabDatasetType.MATCH:
            self.client._set_datalab_matching_dataset(self.data_lab_id, None)
        elif dataset_type == DataLabDatasetType.SEGMENTS:
            self.client._set_datalab_segments_dataset(self.data_lab_id, None)
        else:
            raise Exception(f"Unknown dataset type {dataset_type}")

    def _check_dataset_type_permitted(self, dataset_type: DataLabDatasetType):
        if dataset_type == DataLabDatasetType.EMBEDDINGS:
            if not self.cfg.has_embeddings:
                raise Exception("Embeddings not enabled")
        elif dataset_type == DataLabDatasetType.DEMOGRAPHICS:
            if not self.cfg.has_demographics:
                raise Exception("Demographics not enabled")

    def run(
        self,
        /,
        *,
        dry_run: Optional[DryRunOptions] = None,
        parameters: Optional[Mapping[Text, Text]] = None,
    ):
        """
        Running the DataLab results in the validation jobs and statistics job being kicked off.
        This function does not block waiting for the results. Instead the user should call the
        `get_validation_report` or `get_statistics_report` function.
        """
        features = self._get_features()
        if "COMPUTE_STATISTICS" not in features:
            raise Exception("DataLab does not support computing statistics")

        self._update_enclave_specs()
        self.dcr_id = self._construct_backing_dcr(self.session)
        for dataset_type, dataset in self.datasets.items():
            node_name = self._get_data_lab_node_names(dataset_type, features)
            self.session.publish_dataset(
                self.dcr_id, dataset.manifest_hash, node_name, dataset.key
            )

        # Start validation jobs
        validation_job_id = self.session._submit_compute(
            self.dcr_id,
            self._get_validation_nodes(),
            dry_run=dry_run,
            parameters=parameters,
        ).jobId.hex()

        # Start statistics job
        statistics_job_id = self.session._submit_compute(
            self.dcr_id, ["publisher_data_statistics"]
        ).jobId.hex()

        self.client._set_datalab_job_ids(
            self.data_lab_id,
            validation_job_id,
            statistics_job_id,
            self.session.connection.channel.driver_attestation_specification_hash,
        )

    def get_validation_report(self, timeout: Optional[int] = None):
        """
        Retrieve the validation report. This function will block until the report is ready unless a timeout is specified.

        **Parameters**:
        - `timeout`: Amount of time to wait (in seconds) for the validation report to become available.
        """
        data_lab = self.client.get_data_lab(self.data_lab_id)

        # A validation report is not stored in the DB if statistics have already been produced.
        if data_lab["statistics"]:
            raise Exception(
                "Cannot retrieve validation report. Statistics already calculated."
            )

        driver_attestation_hash = data_lab["jobsDriverAttestationHash"]
        if not driver_attestation_hash:
            raise Exception("Driver attestation spec not found")

        validation_compute_job_id = data_lab["validationComputeJobId"]
        if not validation_compute_job_id:
            raise Exception(
                "Validation compute job ID not found. Please run the DataLab."
            )

        driver_spec = self.client._get_enclave_spec_from_hash(driver_attestation_hash)
        if not driver_spec:
            raise Exception(f"Failed to find driver for data lab {self.data_lab_id}")

        session = create_session_from_driver_spec(self.client, driver_spec)
        validation_nodes = self._get_validation_nodes()
        session.wait_until_computation_has_finished_for_all_compute_nodes(
            validation_compute_job_id, validation_nodes, timeout=timeout
        )
        # Get individual reports
        validation_reports = {}
        for node in validation_nodes:
            job_id = JobId(validation_compute_job_id, node)
            result = session.get_computation_result(job_id, timeout=timeout)
            zip = zipfile.ZipFile(io.BytesIO(result), "r")
            if "validation-report.json" in zip.namelist():
                validation_reports[node.removesuffix("_validation_report")] = (
                    json.loads(zip.read("validation-report.json").decode())
                )
        return validation_reports

    def get_statistics_report(self, timeout: Optional[int] = None):
        """
        Retrieve the statistics report. This function will block until the report is ready unless a timeout is specified.

        **Parameters**:
        - `timeout`: Amount of time to wait (in seconds) for the statistics report to become available.
        """
        features = self._get_features()
        if "COMPUTE_STATISTICS" not in features:
            raise Exception("DataLab does not support computing statistics")

        data_lab = self.client.get_data_lab(self.data_lab_id)
        if data_lab["statistics"]:
            # If we already have the statistics, return them.
            return data_lab["statistics"]

        statistics_compute_job_id = data_lab["statisticsComputeJobId"]
        if not statistics_compute_job_id:
            raise Exception(
                "Statistics compute job ID not found. Please run the DataLab."
            )

        driver_attestation_hash = data_lab["jobsDriverAttestationHash"]
        if not driver_attestation_hash:
            raise Exception("Driver attestation spec not found")

        driver_spec = self.client._get_enclave_spec_from_hash(driver_attestation_hash)
        if not driver_spec:
            raise Exception(f"Failed to find driver for data lab {self.data_lab_id}")

        session = create_session_from_driver_spec(self.client, driver_spec)
        job_id = JobId(statistics_compute_job_id, "publisher_data_statistics")
        results = session.get_computation_result(job_id, interval=5, timeout=timeout)
        zip = zipfile.ZipFile(io.BytesIO(results), "r")
        if "statistics.json" in zip.namelist():
            statistics = zip.read("statistics.json").decode()
            # Store the statistics.
            self.client._set_datalab_statistics(self.data_lab_id, statistics)
            report = json.loads(statistics)
            return report
        else:
            raise Exception("Failed to retrieve statistics")

    def is_validation_passed(self, validation_report: Dict[str, str]) -> bool:
        """
        Check whether or not DataLab validation has passed.

        **Parameters**:
        - `validation_report`: Result of calling `get_validation_report` on this DataLab.
        """
        node_names = self._nodes_to_be_validated()
        for name in node_names:
            if validation_report[name]["report"]["outcome"] != "PASSED":
                return False
        return True

    def _validated(self) -> bool:
        data_lab = self.client.get_data_lab(self.data_lab_id)
        return data_lab["isValidated"]

    def _get_data_lab_datasets_dict(self, data_lab: compiler.DataLab):
        datasets_dict = {}
        for dataset in data_lab["datasets"]:
            datasets_dict[dataset["name"]] = dataset["dataset"]
        return datasets_dict

    # Construct the actual DCR that implements the DataLab functionality.
    def _construct_backing_dcr(self, session: Session) -> str:
        compiled_serialized = compiler.compile_data_lab(self.hl_data_lab)
        low_level_dcr = DataRoom()
        parse_length_delimited(compiled_serialized, low_level_dcr)
        # By default the backing DCR will have the same ID as the DataLab.
        # This causes an issue if there has previously been a DCR created for the DataLab. Instead generate a new ID.
        low_level_dcr.id = str(uuid4())
        dcr_id = session.publish_data_room(
            low_level_dcr, purpose=CreateDcrPurpose.DATA_LAB
        )
        return dcr_id

    # Updates the HL DataLab representation and the session object.
    def _update_enclave_specs(self):
        enclave_specs = get_latest_enclave_specs_as_dictionary(self.client)
        (driver_spec, python_spec) = self._get_data_lab_enclave_specs(enclave_specs)
        # Update to the latest enclave specs
        root_certificate_pem = self.client.decentriq_ca_root_certificate.decode("utf-8")
        # Update the HL representation with the new enclave specs.
        self.hl_data_lab = compiler.update_data_lab_enclave_specifications(
            self.hl_data_lab, driver_spec, python_spec, root_certificate_pem
        )
        auth, _ = self.client.create_auth_using_decentriq_pki(enclave_specs)
        self.session = self.client.create_session(auth, enclave_specs)

    def _get_data_lab_node_names(self, dataset_type: DataLabDatasetType, features: list[str]):
        if dataset_type == DataLabDatasetType.EMBEDDINGS:
            return compiler.get_data_lab_node_id(compiler.DataLabNode.Embeddings, features)
        elif dataset_type == DataLabDatasetType.DEMOGRAPHICS:
            return compiler.get_data_lab_node_id(compiler.DataLabNode.Demographics, features)
        elif dataset_type == DataLabDatasetType.MATCH:
            return compiler.get_data_lab_node_id(compiler.DataLabNode.Users, features)
        elif dataset_type == DataLabDatasetType.SEGMENTS:
            return compiler.get_data_lab_node_id(compiler.DataLabNode.Segments, features)

    def _get_validation_nodes(self):
        validation_nodes = self._nodes_to_be_validated()
        # Add the appropriate suffix for the validation nodes.
        validation_nodes = [node + "_validation_report" for node in validation_nodes]
        return validation_nodes

    # Returns the node names of the nodes that require validation.
    def _nodes_to_be_validated(self) -> list[str]:
        features = self._get_features()
        nodes = []
        if "VALIDATE_MATCHING" in features:
            users = self._get_data_lab_node_names(DataLabDatasetType.MATCH, features)
            nodes.append(users)
        if "VALIDATE_SEGMENTS" in features:
            segments = self._get_data_lab_node_names(DataLabDatasetType.SEGMENTS, features)
            nodes.append(segments)
        if "VALIDATE_EMBEDDINGS" in features and self.cfg.has_embeddings:
            embeddings = self._get_data_lab_node_names(DataLabDatasetType.EMBEDDINGS, features)
            nodes.append(embeddings)
        if "VALIDATE_DEMOGRAPHICS" in features and self.cfg.has_demographics:
            demographics = self._get_data_lab_node_names(
                DataLabDatasetType.DEMOGRAPHICS,
                features
            )
            nodes.append(demographics)
        return nodes

    def _get_features(self):
        features = compiler.get_data_lab_features(self.hl_data_lab)
        return features



from typing import Tuple

from google.protobuf.internal.decoder import _DecodeVarint32  # type: ignore
from google.protobuf.internal.encoder import _VarintBytes  # type: ignore
from google.protobuf.message import Message


# Returns the end offset from serialized_bytes
def parse_length_delimited(
    serialized_bytes: bytes,
    deserialized_object: Message,
    offset: int = 0,
) -> int:
    res: Tuple[int, int] = _DecodeVarint32(serialized_bytes, offset)
    message_length, offset = res
    end_offset = offset + message_length
    deserialized_object.ParseFromString(bytes(serialized_bytes[offset:end_offset]))
    return end_offset


def serialize_length_delimited(message_object: Message) -> bytes:
    serialized: bytes = _VarintBytes(message_object.ByteSize())
    serialized += message_object.SerializeToString()
    return serialized



from __future__ import annotations

import json
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union
from .helper import generate_id, current_iso_standard_utc_time
from typing_extensions import Self
from decentriq_dcr_compiler.schemas import Audience1, Audience2, Audience3


class FilterOperator(str, Enum):
    CONTAINS_ANY = "contains_any_of"
    CONTAINS_NONE = "contains_none_of"
    CONTAINS_ALL = "contains_all_of"
    EMPTY = "empty"
    NOT_EMPTY = "not_empty"


class Filter:
    def __init__(
        self, *, attribute: str, values: List[str], operator: FilterOperator
    ) -> None:
        self.attribute = attribute
        self.values = values
        self.operator = operator

    def as_dict(self) -> Dict[str, Any]:
        return {
            "operator": self.operator.value,
            "attribute": self.attribute,
            "values": self.values,
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Self:
        return cls(
            attribute=d["attribute"],
            values=d["values"],
            operator=FilterOperator(d["operator"]),
        )


class MatchOperator(str, Enum):
    # All filter criteria must be satisfied.
    MATCH_ALL = "and"
    # Any filter criteria may be satisfied.
    MATCH_ANY = "or"


class AudienceFilters:
    def __init__(
        self,
        *,
        filters: List[Filter],
        operator: MatchOperator,
    ) -> None:
        self.filters = filters
        self.operator = operator

    def as_dict(self) -> Dict[str, Any]:
        return {
            "boolean_op": self.operator.value,
            "filters": [f.as_dict() for f in self.filters],
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Self:
        return cls(
            filters=[Filter.from_dict(f) for f in d["filters"]],
            operator=MatchOperator(d["boolean_op"]),
        )


class CombineOperator(str, Enum):
    # Users in both audiences.
    INTERSECT = "intersect"
    # All users.
    UNION = "union"
    # Users in first audience only.
    DIFF = "diff"


class AudienceCombinator:
    def __init__(
        self,
        *,
        operator: CombineOperator,
        source_audience_name: str,
        source_audience_id: str,
        filters: Optional[AudienceFilters] = None,
    ) -> None:
        self.operator = operator
        self.source_ref_name = source_audience_name
        self.source_ref_id = source_audience_id
        self.filters = filters

    def as_dict(self) -> Dict[str, Any]:
        return {
            "operator": self.operator.value,
            "source_ref": self.source_ref_id,
            "filters": self.filters.as_dict() if self.filters else None,
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any], audiences: Dict[str, Any]) -> Self:
        return cls(
            operator=CombineOperator(d["operator"]),
            source_audience_name=audience_name_from_id(d["source_ref"], audiences),
            source_audience_id=d["source_ref"],
            filters=AudienceFilters.from_dict(d["filters"]) if d["filters"] else None,
        )


class AudienceStatus(str, Enum):
    READY = "ready"
    PUBLISHED = "published"
    PUBLISHED_AS_INTERMEDIATE = "published_as_intermediate"


class RuleBasedAudienceDefinition:
    def __init__(
        self,
        *,
        name: str,
        source_ref_name: str,
        source_ref_id: str,
        status: AudienceStatus,
        filters: Optional[AudienceFilters] = None,
        combinators: Optional[List[AudienceCombinator]] = None,
        id: Optional[str] = None,
        created_at: Optional[str] = None,
        audiences: Optional[AudienceDefinitions] = None,
    ) -> None:
        self.name = name
        self.id = id if id else generate_id()
        self.source_ref_name = source_ref_name
        self.source_ref_id = source_ref_id
        self.status = status
        self.filters = filters
        self.combinators = combinators
        self.kind = "rulebased"
        self.created_at = created_at if created_at else current_iso_standard_utc_time()

    def _as_ddc_audience(self) -> Audience3:
        return Audience3.parse_raw(json.dumps(self.as_dict()))

    def as_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "kind": self.kind,
            "source_ref": self.source_ref_id,
            "filters": self.filters.as_dict() if self.filters else None,
            "combine": (
                [c.as_dict() for c in self.combinators] if self.combinators else None
            ),
            "mutable": {
                "name": self.name,
                "status": self.status.value,
                "created_at": self.created_at,
            },
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any], audiences: Dict[str, Any]) -> Self:
        (name, status, created_at) = get_mutable_variables(d["mutable"])
        return cls(
            name=name,
            source_ref_name=audience_name_from_id(d["source_ref"], audiences),
            source_ref_id=d["source_ref"],
            status=status,
            filters=AudienceFilters.from_dict(d["filters"]) if d["filters"] else None,
            combinators=(
                [
                    AudienceCombinator.from_dict(c, audiences=audiences)
                    for c in d["combine"]
                ]
                if d["combine"]
                else None
            ),
            id=d["id"],
            created_at=created_at,
        )


class LookalikeAudienceDefinition:
    def __init__(
        self,
        *,
        name: str,
        reach: int,
        seed_audience_name: str,
        source_ref_id: str,
        exclude_seed_audience: bool,
        status: AudienceStatus,
        id: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> None:
        self.name = name
        self.id = id if id else generate_id()
        self.reach = reach
        self.source_ref_name = seed_audience_name
        self.exclude_seed_audience = exclude_seed_audience
        self.status = status
        self.source_ref_id = source_ref_id
        self.kind = "lookalike"
        self.created_at = created_at if created_at else current_iso_standard_utc_time()

    def _as_ddc_audience(self) -> Audience2:
        return Audience2.parse_raw(json.dumps(self.as_dict()))

    def as_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "kind": self.kind,
            "source_ref": self.source_ref_id,
            "reach": self.reach,
            "exclude_seed_audience": self.exclude_seed_audience,
            "mutable": {
                "name": self.name,
                "status": self.status,
                "created_at": self.created_at,
            },
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any], audiences: Dict[str, Any]) -> Self:
        (name, status, created_at) = get_mutable_variables(d["mutable"])
        return cls(
            name=name,
            reach=d["reach"],
            seed_audience_name=audience_name_from_id(d["source_ref"], audiences),
            source_ref_id=d["source_ref"],
            exclude_seed_audience=d["exclude_seed_audience"],
            status=status,
            id=d["id"],
            created_at=created_at,
        )


class AdvertiserAudienceDefinition:
    def __init__(
        self,
        *,
        name: str,
        audience_size: int,
        audience_type: str,
        status: AudienceStatus,
        id: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> None:
        self.name = name
        self.audience_size = audience_size
        self.audience_type = audience_type
        self.id = id if id else generate_id()
        self.kind = "advertiser"
        self.status = status
        self.created_at = created_at if created_at else current_iso_standard_utc_time()

    def _as_ddc_audience(self) -> Audience1:
        return Audience1.parse_raw(json.dumps(self.as_dict()))

    def as_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "kind": self.kind,
            "audience_size": self.audience_size,
            "audience_type": self.audience_type,
            "mutable": {
                "name": self.name,
                "status": self.status,
                "created_at": self.created_at,
            },
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any], audiences: Dict[str, Any]) -> Self:
        (name, status, created_at) = get_mutable_variables(d["mutable"])
        return cls(
            name=name,
            audience_size=d["audience_size"],
            audience_type=d["audience_type"],
            status=status,
            id=d["id"],
            created_at=created_at,
        )


def audience_name_from_id(audience_id: str, audiences: Dict[str, Any]) -> str:
    lookup = {a["id"]: a["mutable"]["name"] for a in audiences}
    if audience_id not in lookup:
        raise Exception(f'Audience with ID "{audience_id}" could not be found')
    return lookup[audience_id]


def audience_id_from_name(name: str, audiences: AudienceDefinitions) -> str:
    audience = [a for a in audiences if a.name == name]
    if len(audience) == 0:
        raise Exception(
            f'Audience with name "{name}" does not exist in the audience list'
        )
    elif len(audience) > 1:
        raise Exception(f'More than one audience has name "{name}"')
    return audience[0].id


def get_mutable_variables(mutable: Dict[str, Any]) -> Tuple[str, AudienceStatus, str]:
    name = mutable["name"]
    status = AudienceStatus(mutable["status"])
    created_at = mutable["created_at"] if "created_at" in mutable else None
    return (name, status, created_at)


AudienceDefinitions = List[
    Union[
        RuleBasedAudienceDefinition
        , LookalikeAudienceDefinition
        , AdvertiserAudienceDefinition
    ]
]



from typing import Dict, List

from google.protobuf.json_format import MessageToDict

from .client import Client
from .proto import AttestationSpecification, DataRoomConfiguration
from .session import LATEST_GCG_PROTOCOL_VERSION, Session
from .types import EnclaveSpecification


def decode_compute_node_config(
    config: bytes,
    attestation_spec: AttestationSpecification,
    enclave_specs: List[EnclaveSpecification],
):
    for enclave_spec in enclave_specs:
        if enclave_spec["proto"] == attestation_spec:
            decoder = enclave_spec["decoder"]
            if decoder:
                config_decoded = decoder.decode(config)
                return config_decoded
    return None


def decode_compute_node_configs(
    data_room_configuration: DataRoomConfiguration,
    enclave_specs: List[EnclaveSpecification],
) -> Dict[str, Dict]:
    current_configuration = {}
    for element in data_room_configuration.elements:
        current_configuration[element.id] = element
    decoded_configuration = {}

    for element in data_room_configuration.elements:
        if element.HasField("computeNode"):
            if element.computeNode.HasField("branch"):
                config = element.computeNode.branch.config
                index = element.computeNode.branch.attestationSpecificationId
                attestation_spec = current_configuration[index].attestationSpecification
                config_decoded = decode_compute_node_config(
                    config, attestation_spec, enclave_specs
                )
                if config_decoded:
                    decoded_configuration[element.id] = config_decoded

    return decoded_configuration


def decode_data_room_configuration(
    data_room_configuration: DataRoomConfiguration,
    enclave_specifications: List[EnclaveSpecification],
):
    decoded_configs = decode_compute_node_configs(
        data_room_configuration, enclave_specifications
    )
    decoded_configuration = {}

    for element in data_room_configuration.elements:
        element_json = MessageToDict(element)
        is_branch = (
            "computeNode" in element_json and "branch" in element_json["computeNode"]
        )
        if is_branch and element.id in decoded_configs:
            decoded_config = decoded_configs[element.id]
            if decoded_config:
                element_json["computeNode"]["branch"]["config"] = decoded_config
        decoded_configuration[element.id] = element_json

    return decoded_configuration


def get_latest_enclave_specs_as_dictionary(
    client: Client,
) -> Dict[str, EnclaveSpecification]:
    enclave_specs = client._get_enclave_specifications()
    latest_driver_version = 0
    latest_python_worker_version = 0
    latest_enclaves = {}
    for spec in enclave_specs:
        if spec["name"] == "decentriq.driver":
            version = int(spec["version"])
            if version >= latest_driver_version:
                latest_driver_version = version
                latest_enclaves["decentriq.driver"] = spec
        if spec["name"] == "decentriq.python-ml-worker-32-64":
            version = int(spec["version"])
            if version >= latest_python_worker_version:
                latest_python_worker_version = version
                latest_enclaves["decentriq.python-ml-worker-32-64"] = spec
    latest_enclaves["decentriq.driver"]["clientProtocols"] = [
        LATEST_GCG_PROTOCOL_VERSION
    ]
    return latest_enclaves


def create_session_from_driver_spec(
    client: Client, driver_spec: EnclaveSpecification
) -> Session:
    driver_spec["clientProtocols"] = [LATEST_GCG_PROTOCOL_VERSION]
    enclave_specs = {driver_spec["name"]: driver_spec}
    auth, _ = client.create_auth_using_decentriq_pki(enclave_specs)
    session = client.create_session(auth, enclave_specs)
    return session
